from functools import wraps


class disconnect_signal_receiver:
    """
    A decorator that disconnects a signal's receiver during the
    execution of a test and reconnects it back at its end.
    """

    def __init__(self, signal, receiver):
        self.signal = signal
        self.receiver = receiver

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            exception = None
            self.signal.disconnect(self.receiver)

            try:
                func(*args, **kwargs)
            except Exception as e:  # noqa: BLE001
                exception = e
            finally:
                self.signal.connect(self.receiver)

            if exception:
                raise exception

        return wrapper



import logging

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey, GenericRelation
from django.contrib.contenttypes.models import ContentType
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.db.models import Q
from django.db.models.expressions import OuterRef, Subquery
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from modelcluster.models import (
    get_serializable_data_for_fields,
    model_from_serializable_data,
)

from wagtail.log_actions import log
from wagtail.utils.timestamps import ensure_utc

from .content_types import get_default_page_content_type
from .i18n import TranslatableMixin

logger = logging.getLogger("wagtail")


class RevisionQuerySet(models.QuerySet):
    def page_revisions_q(self):
        return Q(base_content_type=get_default_page_content_type())

    def page_revisions(self):
        return self.filter(self.page_revisions_q())

    def not_page_revisions(self):
        return self.exclude(self.page_revisions_q())

    def for_instance(self, instance):
        try:
            # Use RevisionMixin.get_base_content_type() if available
            return self.filter(
                base_content_type=instance.get_base_content_type(),
                object_id=str(instance.pk),
            )
        except AttributeError:
            # Fallback to ContentType for the model
            return self.filter(
                content_type=ContentType.objects.get_for_model(
                    instance, for_concrete_model=False
                ),
                object_id=str(instance.pk),
            )


class RevisionsManager(models.Manager.from_queryset(RevisionQuerySet)):
    def previous_revision_id_subquery(self, revision_fk_name="revision"):
        """
        Returns a Subquery that can be used to annotate a queryset with the ID
        of the previous revision, based on the revision_fk_name field. Useful
        to avoid N+1 queries when generating comparison links between revisions.

        The logic is similar to ``Revision.get_previous().pk``.
        """
        fk = revision_fk_name
        return Subquery(
            Revision.objects.filter(
                base_content_type_id=OuterRef(f"{fk}__base_content_type_id"),
                object_id=OuterRef(f"{fk}__object_id"),
            )
            .filter(
                Q(
                    created_at=OuterRef(f"{fk}__created_at"),
                    pk__lt=OuterRef(f"{fk}__pk"),
                )
                | Q(created_at__lt=OuterRef(f"{fk}__created_at"))
            )
            .order_by("-created_at", "-pk")
            .values_list("pk", flat=True)[:1]
        )


class PageRevisionsManager(RevisionsManager):
    def get_queryset(self):
        return RevisionQuerySet(self.model, using=self._db).page_revisions()


class Revision(models.Model):
    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, related_name="+"
    )
    base_content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE, related_name="+"
    )
    object_id = models.CharField(
        max_length=255,
        verbose_name=_("object id"),
    )
    created_at = models.DateTimeField(db_index=True, verbose_name=_("created at"))
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_("user"),
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="wagtail_revisions",
    )
    object_str = models.TextField(default="")
    content = models.JSONField(
        verbose_name=_("content JSON"), encoder=DjangoJSONEncoder
    )
    approved_go_live_at = models.DateTimeField(
        verbose_name=_("approved go live at"), null=True, blank=True, db_index=True
    )

    objects = RevisionsManager()
    page_revisions = PageRevisionsManager()

    content_object = GenericForeignKey(
        "content_type", "object_id", for_concrete_model=False
    )

    wagtail_reference_index_ignore = True

    @cached_property
    def base_content_object(self):
        return self.base_content_type.get_object_for_this_type(pk=self.object_id)

    def save(self, user=None, *args, **kwargs):
        # Set default value for created_at to now
        # We cannot use auto_now_add as that will override
        # any value that is set before saving
        if self.created_at is None:
            self.created_at = timezone.now()

        # Set default value for base_content_type to the content_type.
        # Page revisions should set this to the default Page model's content type,
        # but the distinction may not be necessary for models that do not use inheritance.
        if self.base_content_type_id is None:
            self.base_content_type_id = self.content_type_id

        super().save(*args, **kwargs)

        if (
            self.approved_go_live_at is None
            and "update_fields" in kwargs
            and "approved_go_live_at" in kwargs["update_fields"]
        ):
            # Log scheduled revision publish cancellation
            object = self.as_object()
            # go_live_at = kwargs['update_fields'][]
            log(
                instance=object,
                action="wagtail.schedule.cancel",
                data={
                    "revision": {
                        "id": self.id,
                        "created": ensure_utc(self.created_at),
                        "go_live_at": ensure_utc(object.go_live_at)
                        if object.go_live_at
                        else None,
                        "has_live_version": object.live,
                    }
                },
                user=user,
                revision=self,
            )

    def as_object(self):
        return self.content_object.with_content_json(self.content)

    def is_latest_revision(self):
        if self.id is None:
            # special case: a revision without an ID is presumed to be newly-created and is thus
            # newer than any revision that might exist in the database
            return True

        latest_revision_id = (
            Revision.objects.filter(
                base_content_type_id=self.base_content_type_id,
                object_id=self.object_id,
            )
            .order_by("-created_at", "-id")
            .values_list("id", flat=True)
            .first()
        )
        return latest_revision_id == self.id

    def delete(self):
        # Update revision_created fields for comments that reference the current revision, if applicable.

        try:
            next_revision = self.get_next()
        except Revision.DoesNotExist:
            next_revision = None

        if next_revision:
            # move comments created on this revision to the next revision, as they may well still apply if they're unresolved
            self.created_comments.all().update(revision_created=next_revision)

        return super().delete()

    def publish(
        self,
        user=None,
        changed=True,
        log_action=True,
        previous_revision=None,
        skip_permission_checks=False,
    ):
        return self.content_object.publish(
            self,
            user=user,
            changed=changed,
            log_action=log_action,
            previous_revision=previous_revision,
            skip_permission_checks=skip_permission_checks,
        )

    def get_previous(self):
        return self.get_previous_by_created_at(
            base_content_type_id=self.base_content_type_id,
            object_id=self.object_id,
        )

    def get_next(self):
        return self.get_next_by_created_at(
            base_content_type_id=self.base_content_type_id,
            object_id=self.object_id,
        )

    def __str__(self):
        return '"' + str(self.content_object) + '" at ' + str(self.created_at)

    class Meta:
        verbose_name = _("revision")
        verbose_name_plural = _("revisions")
        indexes = [
            models.Index(
                fields=["content_type", "object_id"],
                name="content_object_idx",
            ),
            models.Index(
                fields=["base_content_type", "object_id"],
                name="base_content_object_idx",
            ),
        ]


class RevisionMixin(models.Model):
    """A mixin that allows a model to have revisions."""

    latest_revision = models.ForeignKey(
        "wagtailcore.Revision",
        related_name="+",
        verbose_name=_("latest revision"),
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        editable=False,
    )

    _revisions = GenericRelation(
        "wagtailcore.Revision",
        content_type_field="content_type",
        object_id_field="object_id",
        for_concrete_model=False,
    )
    """
    A default ``GenericRelation`` for the purpose of automatically deleting
    revisions when the object is deleted. This is not used to query the object's
    revisions. Instead, the :meth:`revisions` property is used for that purpose.
    As such, this default relation is considered private.

    This ``GenericRelation`` does not have a
    :attr:`~django.contrib.contenttypes.fields.GenericRelation.related_query_name`,
    so it cannot be used for reverse-related queries from ``Revision`` back to
    this model. If the feature is desired, subclasses can define their own
    ``GenericRelation`` to ``Revision`` with a custom ``related_query_name``.

    .. versionadded:: 7.1
        The default ``GenericRelation`` :attr:`~wagtail.models.RevisionMixin._revisions` was added.
    """

    # An array of additional field names that will not be included when the object is copied.
    default_exclude_fields_in_copy = [
        "latest_revision",
    ]

    @property
    def revisions(self):
        """
        Returns revisions that belong to the object. For non-page models, this
        is done by querying the :class:`~wagtail.models.Revision` model directly
        rather than using a
        :class:`~django.contrib.contenttypes.fields.GenericRelation`, to avoid
        `a known limitation <https://code.djangoproject.com/ticket/31269>`_ in
        Django for models with multi-table inheritance where the relation's
        content type may not match the instance's type.
        """
        return Revision.objects.for_instance(self)

    def get_base_content_type(self):
        parents = self._meta.get_parent_list()
        # Get the last non-abstract parent in the MRO as the base_content_type.
        # Note: for_concrete_model=False means that the model can be a proxy model.
        if parents:
            return ContentType.objects.get_for_model(
                parents[-1], for_concrete_model=False
            )
        # This model doesn't inherit from a non-abstract model,
        # use it as the base_content_type.
        return ContentType.objects.get_for_model(self, for_concrete_model=False)

    def get_content_type(self):
        return ContentType.objects.get_for_model(self, for_concrete_model=False)

    def get_latest_revision(self):
        return self.latest_revision

    def get_latest_revision_as_object(self):
        """
        Returns the latest revision of the object as an instance of the model.
        If no latest revision exists, returns the object itself.
        """
        latest_revision = self.get_latest_revision()
        if latest_revision:
            return latest_revision.as_object()
        return self

    def serializable_data(self):
        try:
            return super().serializable_data()
        except AttributeError:
            return get_serializable_data_for_fields(self)

    @classmethod
    def from_serializable_data(cls, data, check_fks=True, strict_fks=False):
        try:
            return super().from_serializable_data(data, check_fks, strict_fks)
        except AttributeError:
            return model_from_serializable_data(
                cls, data, check_fks=check_fks, strict_fks=strict_fks
            )

    def with_content_json(self, content):
        """
        Returns a new version of the object with field values updated to reflect changes
        in the provided ``content`` (which usually comes from a previously-saved revision).

        Certain field values are preserved in order to prevent errors if the returned
        object is saved, such as ``id``. The following field values are also preserved,
        as they are considered to be meaningful to the object as a whole, rather than
        to a specific revision:

        * ``latest_revision``

        If :class:`~wagtail.models.TranslatableMixin` is applied, the following field values
        are also preserved:

        * ``translation_key``
        * ``locale``
        """
        obj = self.from_serializable_data(content)

        # This should definitely never change between revisions
        obj.pk = self.pk

        # Ensure other values that are meaningful for the object as a whole
        # (rather than to a specific revision) are preserved
        obj.latest_revision = self.latest_revision

        if isinstance(self, TranslatableMixin):
            obj.translation_key = self.translation_key
            obj.locale = self.locale

        return obj

    def _update_from_revision(self, revision, changed=True):
        self.latest_revision = revision
        self.save(update_fields=["latest_revision"])

    def save_revision(
        self,
        user=None,
        approved_go_live_at=None,
        changed=True,
        log_action=False,
        previous_revision=None,
        clean=True,
    ):
        """
        Creates and saves a revision.

        :param user: The user performing the action.
        :param approved_go_live_at: The date and time the revision is approved to go live.
        :param changed: Indicates whether there were any content changes.
        :param log_action: Flag for logging the action. Pass ``True`` to also create a log entry. Can be passed an action string.
            Defaults to ``"wagtail.edit"`` when no ``previous_revision`` param is passed, otherwise ``"wagtail.revert"``.
        :param previous_revision: Indicates a revision reversal. Should be set to the previous revision instance.
        :type previous_revision: Revision
        :param clean: Set this to ``False`` to skip cleaning object content before saving this revision.
        :return: The newly created revision.
        """
        if clean:
            self.full_clean()

        revision = Revision.objects.create(
            content_object=self,
            base_content_type=self.get_base_content_type(),
            user=user,
            approved_go_live_at=approved_go_live_at,
            content=self.serializable_data(),
            object_str=str(self),
        )

        self._update_from_revision(revision, changed)

        logger.info(
            'Edited: "%s" pk=%d revision_id=%d', str(self), self.pk, revision.id
        )
        if log_action:
            if not previous_revision:
                log(
                    instance=self,
                    action=log_action
                    if isinstance(log_action, str)
                    else "wagtail.edit",
                    user=user,
                    revision=revision,
                    content_changed=changed,
                )
            else:
                log(
                    instance=self,
                    action=log_action
                    if isinstance(log_action, str)
                    else "wagtail.revert",
                    user=user,
                    data={
                        "revision": {
                            "id": previous_revision.id,
                            "created": ensure_utc(previous_revision.created_at),
                        }
                    },
                    revision=revision,
                    content_changed=changed,
                )

        return revision

    class Meta:
        abstract = True



from django.contrib.auth.models import Permission
from django.test import TestCase
from django.urls import reverse
from django.utils.http import urlencode

from wagtail.images import get_image_model
from wagtail.images.tests.utils import get_test_image_file
from wagtail.test.testapp.models import VariousOnDeleteModel
from wagtail.test.utils import WagtailTestUtils

Image = get_image_model()
test_file = get_test_image_file()


class TestImageBulkDeleteView(WagtailTestUtils, TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.images = [
            Image.objects.create(title=f"Test image - {i}", file=test_file)
            for i in range(1, 6)
        ]
        cls.url = reverse(
            "wagtail_bulk_action",
            args=(Image._meta.app_label, Image._meta.model_name, "delete"),
        )
        cls.query_params = {
            "next": reverse("wagtailimages:index"),
            "id": [item.pk for item in cls.images],
        }
        cls.url += "?" + urlencode(cls.query_params, doseq=True)

    def setUp(self):
        self.user = self.login()

    def test_delete_with_limited_permissions(self):
        self.user.is_superuser = False
        self.user.user_permissions.add(
            Permission.objects.get(
                content_type__app_label="wagtailadmin", codename="access_admin"
            )
        )
        self.user.save()

        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)

        html = response.content.decode()
        self.assertInHTML(
            "<p>You don't have permission to delete these images</p>", html
        )

        for image in self.images:
            self.assertInHTML(f"<li>{image.title}</li>", html)

        response = self.client.post(self.url)
        # User should be redirected back to the index
        self.assertEqual(response.status_code, 302)

        # Images should not be deleted
        for image in self.images:
            self.assertTrue(Image.objects.filter(id=image.id).exists())

    def test_simple(self):
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(
            response, "wagtailimages/bulk_actions/confirm_bulk_delete.html"
        )

    def test_delete(self):
        # Make post request
        response = self.client.post(self.url)

        # User should be redirected back to the index
        self.assertEqual(response.status_code, 302)

        # Images should be deleted
        for image in self.images:
            self.assertFalse(Image.objects.filter(id=image.id).exists())

    def test_usage_link(self):
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        self.assertTemplateUsed(
            response, "wagtailimages/bulk_actions/confirm_bulk_delete.html"
        )
        for image in self.images:
            self.assertContains(response, image.usage_url)
        # usage count should be printed for each image
        self.assertContains(response, "This image is referenced 0 times.", count=5)

    def test_delete_get_with_protected_reference(self):
        protected = self.images[0]
        with self.captureOnCommitCallbacks(execute=True):
            VariousOnDeleteModel.objects.create(
                text="Undeletable",
                protected_image=protected,
            )
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, 200)
        soup = self.get_soup(response.content)
        main = soup.select_one("main")
        usage_link = main.find(
            "a",
            href=reverse("wagtailimages:image_usage", args=[protected.pk])
            + "?describe_on_delete=1",
        )
        self.assertIsNotNone(usage_link)
        self.assertEqual(usage_link.text.strip(), "This image is referenced 1 time.")
        self.assertContains(
            response,
            "One or more references to this image prevent it from being deleted.",
        )
        submit_button = main.select_one("form button[type=submit]")
        self.assertIsNone(submit_button)
        back_button = main.find("a", href=reverse("wagtailimages:index"))
        self.assertIsNotNone(back_button)
        self.assertEqual(back_button.text.strip(), "Go back")

    def test_delete_post_with_protected_reference(self):
        protected = self.images[0]
        with self.captureOnCommitCallbacks(execute=True):
            VariousOnDeleteModel.objects.create(
                text="Undeletable",
                protected_image=protected,
            )
        response = self.client.post(self.url)

        # Should throw a PermissionDenied error and redirect to the dashboard
        self.assertEqual(response.status_code, 302)
        self.assertRedirects(response, reverse("wagtailadmin_home"))
        self.assertEqual(
            response.context["message"],
            "Sorry, you do not have permission to access this area.",
        )

        # Check that the image is still here
        self.assertTrue(Image.objects.filter(pk=protected.pk).exists())



from django.conf import settings
from django.contrib.auth import views as auth_views
from django.urls import path, re_path

from wagtail import views
from wagtail.coreutils import WAGTAIL_APPEND_SLASH

if WAGTAIL_APPEND_SLASH:
    # If WAGTAIL_APPEND_SLASH is True (the default value), we match a
    # (possibly empty) list of path segments ending in slashes.
    # CommonMiddleware will redirect requests without a trailing slash to
    # a URL with a trailing slash
    serve_pattern = r"^((?:[\w\-]+/)*)$"
else:
    # If WAGTAIL_APPEND_SLASH is False, allow Wagtail to serve pages on URLs
    # with and without trailing slashes
    serve_pattern = r"^([\w\-/]*)$"


WAGTAIL_FRONTEND_LOGIN_TEMPLATE = getattr(
    settings, "WAGTAIL_FRONTEND_LOGIN_TEMPLATE", "wagtailcore/login.html"
)


urlpatterns = [
    path(
        "_util/authenticate_with_password/<int:page_view_restriction_id>/<int:page_id>/",
        views.authenticate_with_password,
        name="wagtailcore_authenticate_with_password",
    ),
    path(
        "_util/login/",
        auth_views.LoginView.as_view(template_name=WAGTAIL_FRONTEND_LOGIN_TEMPLATE),
        name="wagtailcore_login",
    ),
    # Front-end page views are handled through Wagtail's core.views.serve
    # mechanism
    re_path(serve_pattern, views.serve, name="wagtail_serve"),
]



from django.apps import apps
from django.conf import settings
from django.contrib.admin.utils import quote
from django.core import checks
from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.http import Http404
from django.urls import path, re_path, reverse, reverse_lazy
from django.utils.functional import cached_property
from django.utils.text import capfirst
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy

from wagtail import hooks
from wagtail.admin.checks import check_panels_in_model
from wagtail.admin.panels import ObjectList, extract_panel_definitions_from_model_class
from wagtail.admin.ui.components import MediaContainer
from wagtail.admin.ui.menus import MenuItem
from wagtail.admin.ui.side_panels import ChecksSidePanel, PreviewSidePanel
from wagtail.admin.ui.tables import (
    BulkActionsCheckboxColumn,
    NumberColumn,
    TitleColumn,
)
from wagtail.admin.views import generic
from wagtail.admin.views.generic import history, lock, workflow
from wagtail.admin.views.generic.permissions import PermissionCheckedMixin
from wagtail.admin.views.generic.preview import (
    PreviewOnCreate,
    PreviewOnEdit,
    PreviewRevision,
)
from wagtail.admin.viewsets import viewsets
from wagtail.admin.viewsets.model import ModelViewSet, ModelViewSetGroup
from wagtail.admin.widgets.button import (
    BaseButton,
    Button,
    ButtonWithDropdown,
)
from wagtail.models import (
    DraftStateMixin,
    LockableMixin,
    PreviewableMixin,
    RevisionMixin,
    WorkflowMixin,
)
from wagtail.permissions import ModelPermissionPolicy
from wagtail.snippets.action_menu import SnippetActionMenu
from wagtail.snippets.models import SnippetAdminURLFinder, get_snippet_models
from wagtail.snippets.side_panels import SnippetStatusSidePanel
from wagtail.snippets.views.chooser import SnippetChooserViewSet


# == Helper functions ==
def get_snippet_model_from_url_params(app_name, model_name):
    """
    Retrieve a model from an app_label / model_name combo.
    Raise Http404 if the model is not a valid snippet type.
    """
    try:
        model = apps.get_model(app_name, model_name)
    except LookupError:
        raise Http404
    if model not in get_snippet_models():
        # don't allow people to hack the URL to edit content types that aren't registered as snippets
        raise Http404

    return model


# == Views ==


def get_snippet_models_for_index_view():
    models = get_snippet_models()

    if getattr(settings, "WAGTAILSNIPPETS_MENU_SHOW_ALL", False):
        return models

    return [
        model
        for model in models
        if not model.snippet_viewset.get_menu_item_is_registered()
    ]


class ModelIndexView(generic.BaseListingView):
    page_title = gettext_lazy("Snippets")
    header_icon = "snippet"
    index_url_name = "wagtailsnippets:index"
    default_ordering = "name"

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.snippet_types = self._get_snippet_types()

    def _get_snippet_types(self):
        return [
            {
                "name": capfirst(model._meta.verbose_name_plural),
                "count": model._default_manager.all().count(),
                "model": model,
                "url": url,
            }
            for model in get_snippet_models_for_index_view()
            if (url := self.get_list_url(model))
        ]

    def dispatch(self, request, *args, **kwargs):
        if not self.snippet_types:
            raise PermissionDenied
        return super().dispatch(request, *args, **kwargs)

    def get_list_url(self, model):
        if model.snippet_viewset.permission_policy.user_has_any_permission(
            self.request.user,
            {"add", "change", "delete", "view"},
        ):
            return reverse(model.snippet_viewset.get_url_name("list"))

    def get_queryset(self):
        return None

    @cached_property
    def columns(self):
        return [
            TitleColumn(
                "name",
                label=_("Name"),
                get_url=lambda type: type["url"],
                sort_key="name",
            ),
            NumberColumn(
                "count",
                label=_("Instances"),
                sort_key="count",
            ),
        ]

    def get_context_data(self, **kwargs):
        reverse = self.ordering[0] == "-"

        if self.ordering in ["count", "-count"]:
            snippet_types = sorted(
                self.snippet_types,
                key=lambda type: type["count"],
                reverse=reverse,
            )
        else:
            snippet_types = sorted(
                self.snippet_types,
                key=lambda type: type["name"].lower(),
                reverse=reverse,
            )

        return super().get_context_data(object_list=snippet_types)

    def get_template_names(self):
        # We use the generic index template instead of model_index.html,
        # but we look for it anyway so users can customise this view's template
        # without having to override the entire view or the generic template.
        return [
            "wagtailsnippets/snippets/model_index.html",
            self.template_name,
        ]


class IndexView(generic.IndexViewOptionalFeaturesMixin, generic.IndexView):
    view_name = "list"

    def get_base_queryset(self):
        # Allow the queryset to be a callable that takes a request
        # so that it can be evaluated in the context of the request
        if callable(self.queryset):
            self.queryset = self.queryset(self.request)
        return super().get_base_queryset()

    @cached_property
    def columns(self):
        return [
            BulkActionsCheckboxColumn("bulk_actions", obj_type="snippet"),
            *super().columns,
        ]

    def _get_title_column(self, *args, **kwargs):
        return super()._get_title_column(
            *args,
            **kwargs,
            get_title_id=lambda instance: f"snippet_{quote(instance.pk)}_title",
        )

    def get_list_buttons(self, instance):
        next_url = self.request.path
        list_buttons = []
        more_buttons = []

        buttons = self.get_list_more_buttons(instance)
        for hook in hooks.get_hooks("register_snippet_listing_buttons"):
            buttons.extend(hook(instance, self.request.user, next_url))

        for button in buttons:
            if isinstance(button, BaseButton) and not button.allow_in_dropdown:
                # If the button is not allowed in a dropdown menu, add it to
                # the top-level list of buttons
                list_buttons.append(button)
            elif isinstance(button, MenuItem):
                # Allow simple MenuItem instances to be passed in directly
                if button.is_shown(self.request.user):
                    more_buttons.append(Button.from_menu_item(button))
            elif button.show:
                # Otherwise, add it to the default "More" dropdown
                more_buttons.append(button)

        # Pass the more_buttons to the construct hooks, as that's what contains
        # the default buttons and most buttons added via register_snippet_listing_buttons
        for hook in hooks.get_hooks("construct_snippet_listing_buttons"):
            hook(more_buttons, instance, self.request.user)

        if more_buttons:
            list_buttons.append(
                ButtonWithDropdown(
                    buttons=more_buttons,
                    icon_name="dots-horizontal",
                    attrs={
                        "aria-label": _("More options for '%(title)s'")
                        % {"title": str(instance)},
                    },
                )
            )

        return list_buttons


class CreateView(generic.CreateEditViewOptionalFeaturesMixin, generic.CreateView):
    view_name = "create"
    template_name = "wagtailsnippets/snippets/create.html"

    def run_before_hook(self):
        return self.run_hook("before_create_snippet", self.request, self.model)

    def run_after_hook(self):
        return self.run_hook("after_create_snippet", self.request, self.object)

    def _get_action_menu(self):
        return SnippetActionMenu(self.request, view=self.view_name, model=self.model)

    def get_side_panels(self):
        side_panels = [
            SnippetStatusSidePanel(
                self.form.instance,
                self.request,
                show_schedule_publishing_toggle=getattr(
                    self.form, "show_schedule_publishing_toggle", False
                ),
                locale=self.locale,
                translations=self.translations,
            )
        ]
        if self.preview_enabled and self.form.instance.is_previewable():
            side_panels.append(
                PreviewSidePanel(
                    self.form.instance, self.request, preview_url=self.get_preview_url()
                )
            )
            side_panels.append(ChecksSidePanel(self.form.instance, self.request))
        return MediaContainer(side_panels)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        action_menu = self._get_action_menu()
        context["media"] += action_menu.media
        context["action_menu"] = action_menu
        return context


class CopyView(generic.CopyViewMixin, CreateView):
    pass


class EditView(generic.CreateEditViewOptionalFeaturesMixin, generic.EditView):
    view_name = "edit"
    template_name = "wagtailsnippets/snippets/edit.html"

    def run_before_hook(self):
        return self.run_hook("before_edit_snippet", self.request, self.object)

    def run_after_hook(self):
        return self.run_hook("after_edit_snippet", self.request, self.object)

    def _get_action_menu(self):
        return SnippetActionMenu(
            self.request,
            view=self.view_name,
            instance=self.object,
            locked_for_user=self.locked_for_user,
        )

    def get_side_panels(self):
        side_panels = [
            SnippetStatusSidePanel(
                self.object,
                self.request,
                show_schedule_publishing_toggle=getattr(
                    self.form, "show_schedule_publishing_toggle", False
                ),
                live_object=self.live_object,
                scheduled_object=self.live_object.get_scheduled_revision_as_object()
                if self.draftstate_enabled
                else None,
                locale=self.locale,
                translations=self.translations,
                usage_url=self.get_usage_url(),
                history_url=self.get_history_url(),
                last_updated_info=self.get_last_updated_info(),
            )
        ]
        if self.preview_enabled and self.object.is_previewable():
            side_panels.append(
                PreviewSidePanel(
                    self.object, self.request, preview_url=self.get_preview_url()
                )
            )
            side_panels.append(ChecksSidePanel(self.object, self.request))
        return MediaContainer(side_panels)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        action_menu = self._get_action_menu()
        context["media"] += action_menu.media
        context["action_menu"] = action_menu
        return context


class DeleteView(generic.DeleteView):
    view_name = "delete"

    def run_before_hook(self):
        return self.run_hook("before_delete_snippet", self.request, [self.object])

    def run_after_hook(self):
        return self.run_hook("after_delete_snippet", self.request, [self.object])


class UsageView(generic.UsageView):
    view_name = "usage"


class HistoryView(history.HistoryView):
    view_name = "history"


class InspectView(generic.InspectView):
    view_name = "inspect"


class PreviewOnCreateView(PreviewOnCreate):
    pass


class PreviewOnEditView(PreviewOnEdit):
    pass


class PreviewRevisionView(PermissionCheckedMixin, PreviewRevision):
    permission_required = "change"


class RevisionsCompareView(PermissionCheckedMixin, generic.RevisionsCompareView):
    permission_required = "change"


class UnpublishView(PermissionCheckedMixin, generic.UnpublishView):
    permission_required = "publish"


class RevisionsUnscheduleView(PermissionCheckedMixin, generic.RevisionsUnscheduleView):
    permission_required = "publish"


class LockView(PermissionCheckedMixin, lock.LockView):
    permission_required = "lock"

    def user_has_permission(self, permission):
        if self.request.user.is_superuser:
            return True

        if permission == self.permission_required and isinstance(
            self.object, WorkflowMixin
        ):
            current_workflow_task = self.object.current_workflow_task
            if current_workflow_task:
                return current_workflow_task.user_can_lock(
                    self.object, self.request.user
                )

        return super().user_has_permission(permission)


class UnlockView(PermissionCheckedMixin, lock.UnlockView):
    permission_required = "unlock"

    def user_has_permission(self, permission):
        if self.request.user.is_superuser:
            return True

        if permission == self.permission_required:
            # Allow unlocking even if the user does not have the 'unlock' permission
            # if they are the user who locked the object
            if self.object.locked_by_id == self.request.user.pk:
                return True

            if isinstance(self.object, WorkflowMixin):
                current_workflow_task = self.object.current_workflow_task
                if current_workflow_task:
                    return current_workflow_task.user_can_unlock(
                        self.object, self.request.user
                    )

        return super().user_has_permission(permission)


class WorkflowActionView(workflow.WorkflowAction):
    pass


class CollectWorkflowActionDataView(workflow.CollectWorkflowActionData):
    pass


class ConfirmWorkflowCancellationView(workflow.ConfirmWorkflowCancellation):
    pass


class WorkflowPreviewView(workflow.PreviewRevisionForTask):
    pass


class WorkflowHistoryView(PermissionCheckedMixin, history.WorkflowHistoryView):
    permission_required = "change"


class WorkflowHistoryDetailView(
    PermissionCheckedMixin, history.WorkflowHistoryDetailView
):
    permission_required = "change"


class SnippetViewSet(ModelViewSet):
    """
    A viewset that instantiates the admin views for snippets.

    All attributes and methods from
    :class:`~wagtail.admin.viewsets.model.ModelViewSet` are available.

    For more information on how to use this class,
    see :ref:`wagtailsnippets_custom_admin_views`.
    """

    #: The model class to be registered as a snippet with this viewset.
    model = None

    #: The number of items to display in the chooser view. Defaults to 10.
    chooser_per_page = 10

    #: The URL namespace to use for the admin views.
    #: If left unset, ``wagtailsnippets_{app_label}_{model_name}`` is used instead.
    #:
    #: **Deprecated** - the preferred attribute to customise is :attr:`~.ViewSet.url_namespace`.
    admin_url_namespace = None

    #: The base URL path to use for the admin views.
    #: If left unset, ``snippets/{app_label}/{model_name}`` is used instead.
    #:
    #: **Deprecated** - the preferred attribute to customise is :attr:`~.ViewSet.url_prefix`.
    base_url_path = None

    #: The URL namespace to use for the chooser admin views.
    #: If left unset, ``wagtailsnippetchoosers_{app_label}_{model_name}`` is used instead.
    chooser_admin_url_namespace = None

    #: The base URL path to use for the chooser admin views.
    #: If left unset, ``snippets/choose/{app_label}/{model_name}`` is used instead.
    chooser_base_url_path = None

    #: The view class to use for the index view; must be a subclass of ``wagtail.snippets.views.snippets.IndexView``.
    index_view_class = IndexView

    #: The view class to use for the create view; must be a subclass of ``wagtail.snippets.views.snippets.CreateView``.
    add_view_class = CreateView

    #: The view class to use for the copy view; must be a subclass of ``wagtail.snippet.views.snippets.CopyView``.
    copy_view_class = CopyView

    #: The view class to use for the edit view; must be a subclass of ``wagtail.snippets.views.snippets.EditView``.
    edit_view_class = EditView

    #: The view class to use for the delete view; must be a subclass of ``wagtail.snippets.views.snippets.DeleteView``.
    delete_view_class = DeleteView

    #: The view class to use for the usage view; must be a subclass of ``wagtail.snippets.views.snippets.UsageView``.
    usage_view_class = UsageView

    #: The view class to use for the history view; must be a subclass of ``wagtail.snippets.views.snippets.HistoryView``.
    history_view_class = HistoryView

    #: The view class to use for the inspect view; must be a subclass of ``wagtail.snippets.views.snippets.InspectView``.
    inspect_view_class = InspectView

    #: The view class to use for previewing revisions; must be a subclass of ``wagtail.snippets.views.snippets.PreviewRevisionView``.
    revisions_view_class = PreviewRevisionView

    #: The view class to use for comparing revisions; must be a subclass of ``wagtail.snippets.views.snippets.RevisionsCompareView``.
    revisions_compare_view_class = RevisionsCompareView

    #: The view class to use for unscheduling revisions; must be a subclass of ``wagtail.snippets.views.snippets.RevisionsUnscheduleView``.
    revisions_unschedule_view_class = RevisionsUnscheduleView

    #: The view class to use for unpublishing a snippet; must be a subclass of ``wagtail.snippets.views.snippets.UnpublishView``.
    unpublish_view_class = UnpublishView

    #: The view class to use for previewing on the create view; must be a subclass of ``wagtail.snippets.views.snippets.PreviewOnCreateView``.
    preview_on_add_view_class = PreviewOnCreateView

    #: The view class to use for previewing on the edit view; must be a subclass of ``wagtail.snippets.views.snippets.PreviewOnEditView``.
    preview_on_edit_view_class = PreviewOnEditView

    #: The view class to use for locking a snippet; must be a subclass of ``wagtail.snippets.views.snippets.LockView``.
    lock_view_class = LockView

    #: The view class to use for unlocking a snippet; must be a subclass of ``wagtail.snippets.views.snippets.UnlockView``.
    unlock_view_class = UnlockView

    #: The view class to use for performing a workflow action; must be a subclass of ``wagtail.snippets.views.snippets.WorkflowActionView``.
    workflow_action_view_class = WorkflowActionView

    #: The view class to use for performing a workflow action that returns the validated data in the response; must be a subclass of ``wagtail.snippets.views.snippets.CollectWorkflowActionDataView``.
    collect_workflow_action_data_view_class = CollectWorkflowActionDataView

    #: The view class to use for confirming the cancellation of a workflow; must be a subclass of ``wagtail.snippets.views.snippets.ConfirmWorkflowCancellationView``.
    confirm_workflow_cancellation_view_class = ConfirmWorkflowCancellationView

    #: The view class to use for previewing a revision for a specific task; must be a subclass of ``wagtail.snippets.views.snippets.WorkflowPreviewView``.
    workflow_preview_view_class = WorkflowPreviewView

    #: The view class to use for the workflow history view; must be a subclass of ``wagtail.snippets.views.snippets.WorkflowHistoryView``.
    workflow_history_view_class = WorkflowHistoryView

    #: The view class to use for the workflow history detail view; must be a subclass of ``wagtail.snippets.views.snippets.WorkflowHistoryDetailView``.
    workflow_history_detail_view_class = WorkflowHistoryDetailView

    #: The ViewSet class to use for the chooser views; must be a subclass of ``wagtail.snippets.views.chooser.SnippetChooserViewSet``.
    chooser_viewset_class = SnippetChooserViewSet

    #: The prefix of template names to look for when rendering the admin views.
    template_prefix = "wagtailsnippets/snippets/"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        if self.model is None:
            raise ImproperlyConfigured(
                "SnippetViewSet must be passed a model or define a model attribute."
            )

        self.model_opts = self.model._meta
        self.app_label = self.model_opts.app_label
        self.model_name = self.model_opts.model_name

        self.preview_enabled = issubclass(self.model, PreviewableMixin)
        self.revision_enabled = issubclass(self.model, RevisionMixin)
        self.draftstate_enabled = issubclass(self.model, DraftStateMixin)
        self.workflow_enabled = issubclass(self.model, WorkflowMixin)
        self.locking_enabled = issubclass(self.model, LockableMixin)

        self.menu_item_is_registered = getattr(
            self, "menu_item_is_registered", bool(self.menu_hook)
        )

    @cached_property
    def url_prefix(self):
        # SnippetViewSet historically allows overriding the URL prefix via the
        # get_admin_base_path method or the admin_base_path attribute, so preserve that here
        return self.get_admin_base_path()

    @cached_property
    def url_namespace(self):
        # SnippetViewSet historically allows overriding the URL namespace via the
        # get_admin_url_namespace method or the admin_url_namespace attribute,
        # so preserve that here
        return self.get_admin_url_namespace()

    @property
    def revisions_revert_view_class(self):
        """
        The view class to use for reverting to a previous revision.

        By default, this class is generated by combining the edit view with
        ``wagtail.admin.views.generic.mixins.RevisionsRevertMixin``. As a result,
        this class must be a subclass of ``wagtail.snippets.views.snippets.EditView``
        and must handle the reversion correctly.
        """
        revisions_revert_view_class = type(
            "_RevisionsRevertView",
            (generic.RevisionsRevertMixin, self.edit_view_class),
            {"view_name": "revisions_revert"},
        )
        return revisions_revert_view_class

    @property
    def permission_policy(self):
        return ModelPermissionPolicy(self.model)

    def get_common_view_kwargs(self, **kwargs):
        return super().get_common_view_kwargs(
            **{
                "index_url_name": self.get_url_name("list"),
                "index_results_url_name": self.get_url_name("list_results"),
                "lock_url_name": self.get_url_name("lock"),
                "unlock_url_name": self.get_url_name("unlock"),
                "revisions_view_url_name": self.get_url_name("revisions_view"),
                "revisions_revert_url_name": self.get_url_name("revisions_revert"),
                "revisions_compare_url_name": self.get_url_name("revisions_compare"),
                "revisions_unschedule_url_name": self.get_url_name(
                    "revisions_unschedule"
                ),
                "unpublish_url_name": self.get_url_name("unpublish"),
                "breadcrumbs_items": self.breadcrumbs_items,
                **kwargs,
            }
        )

    def get_index_view_kwargs(self, **kwargs):
        return super().get_index_view_kwargs(
            queryset=self.get_queryset,
            **kwargs,
        )

    def get_add_view_kwargs(self, **kwargs):
        return super().get_add_view_kwargs(
            preview_url_name=self.get_url_name("preview_on_add"),
            **kwargs,
        )

    def get_copy_view_kwargs(self, **kwargs):
        return self.get_add_view_kwargs(**kwargs)

    def get_edit_view_kwargs(self, **kwargs):
        return super().get_edit_view_kwargs(
            preview_url_name=self.get_url_name("preview_on_edit"),
            workflow_history_url_name=self.get_url_name("workflow_history"),
            confirm_workflow_cancellation_url_name=self.get_url_name(
                "confirm_workflow_cancellation"
            ),
            **kwargs,
        )

    @property
    def revisions_view(self):
        return self.construct_view(self.revisions_view_class)

    @property
    def revisions_revert_view(self):
        return self.construct_view(
            self.revisions_revert_view_class,
            **self.get_edit_view_kwargs(),
        )

    @property
    def revisions_compare_view(self):
        return self.construct_view(
            self.revisions_compare_view_class,
            edit_handler=self._edit_handler,
            template_name=self.get_templates(
                "revisions_compare",
                fallback=self.revisions_compare_view_class.template_name,
            ),
        )

    @property
    def revisions_unschedule_view(self):
        return self.construct_view(
            self.revisions_unschedule_view_class,
            template_name=self.get_templates(
                "revisions_unschedule",
                fallback=self.revisions_unschedule_view_class.template_name,
            ),
        )

    @property
    def unpublish_view(self):
        return self.construct_view(
            self.unpublish_view_class,
            template_name=self.get_templates(
                "unpublish", fallback=self.unpublish_view_class.template_name
            ),
        )

    @property
    def preview_on_add_view(self):
        return self.construct_view(
            self.preview_on_add_view_class,
            form_class=self.get_form_class(),
        )

    @property
    def preview_on_edit_view(self):
        return self.construct_view(
            self.preview_on_edit_view_class,
            form_class=self.get_form_class(for_update=True),
        )

    @property
    def lock_view(self):
        return self.construct_view(
            self.lock_view_class,
            success_url_name=self.get_url_name("edit"),
        )

    @property
    def copy_view(self):
        return self.construct_view(self.copy_view_class, **self.get_copy_view_kwargs())

    @property
    def unlock_view(self):
        return self.construct_view(
            self.unlock_view_class,
            success_url_name=self.get_url_name("edit"),
        )

    @property
    def workflow_action_view(self):
        return self.construct_view(
            self.workflow_action_view_class,
            redirect_url_name=self.get_url_name("edit"),
            submit_url_name=self.get_url_name("workflow_action"),
        )

    @property
    def collect_workflow_action_data_view(self):
        return self.construct_view(
            self.collect_workflow_action_data_view_class,
            redirect_url_name=self.get_url_name("edit"),
            submit_url_name=self.get_url_name("collect_workflow_action_data"),
        )

    @property
    def confirm_workflow_cancellation_view(self):
        return self.construct_view(self.confirm_workflow_cancellation_view_class)

    @property
    def workflow_preview_view(self):
        return self.construct_view(self.workflow_preview_view_class)

    @property
    def workflow_history_view(self):
        return self.construct_view(
            self.workflow_history_view_class,
            template_name=self.get_templates(
                "workflow_history/index",
                fallback=self.workflow_history_view_class.template_name,
            ),
            workflow_history_detail_url_name=self.get_url_name(
                "workflow_history_detail"
            ),
        )

    @property
    def workflow_history_detail_view(self):
        return self.construct_view(
            self.workflow_history_detail_view_class,
            template_name=self.get_templates(
                "workflow_history/detail",
                fallback=self.workflow_history_detail_view_class.template_name,
            ),
            workflow_history_url_name=self.get_url_name("workflow_history"),
        )

    @property
    def chooser_viewset(self):
        return self.chooser_viewset_class(
            self.get_chooser_admin_url_namespace(),
            model=self.model,
            url_prefix=self.get_chooser_admin_base_path(),
            icon=self.icon,
            per_page=self.chooser_per_page,
        )

    @cached_property
    def icon(self):
        return self.get_icon()

    def get_icon(self):
        """
        Returns the icon to be used for the admin views.

        **Deprecated** - the preferred way to customise this is to define an ``icon`` property.
        """
        return "snippet"

    @cached_property
    def menu_label(self):
        return self.get_menu_label()

    def get_menu_label(self):
        """
        Returns the label text to be used for the menu item.

        **Deprecated** - the preferred way to customise this is to define a ``menu_label`` property.
        """
        return capfirst(self.model_opts.verbose_name_plural)

    @cached_property
    def menu_name(self):
        return self.get_menu_name()

    def get_menu_name(self):
        """
        Returns the name to be used for the menu item.

        **Deprecated** - the preferred way to customise this is to define a ``menu_name`` property.
        """
        return ""

    @cached_property
    def menu_icon(self):
        return self.get_menu_icon()

    def get_menu_icon(self):
        """
        Returns the icon to be used for the menu item.

        **Deprecated** - the preferred way to customise this is to define a ``menu_icon`` property.
        """
        return self.icon

    @cached_property
    def menu_order(self):
        return self.get_menu_order()

    def get_menu_order(self):
        """
        Returns the ordering number to be applied to the menu item.

        **Deprecated** - the preferred way to customise this is to define a ``menu_order`` property.
        """
        # By default, put it at the last item before Reports, whose order is 9000.
        return 8999

    def get_menu_item_is_registered(self):
        return self.menu_item_is_registered

    @property
    def breadcrumbs_items(self):
        # Use reverse_lazy instead of reverse
        # because this will be passed to the view classes at startup
        breadcrumbs = [
            {"url": reverse_lazy("wagtailadmin_home"), "label": _("Home")},
        ]

        if (
            getattr(settings, "WAGTAILSNIPPETS_MENU_SHOW_ALL", False)
            or not self.get_menu_item_is_registered()
        ):
            breadcrumbs.append(
                {"url": reverse_lazy("wagtailsnippets:index"), "label": _("Snippets")},
            )

        return breadcrumbs

    def get_queryset(self, request):
        """
        Returns a QuerySet of all model instances to be shown on the index view.
        If ``None`` is returned (the default), the logic in
        ``index_view.get_base_queryset()`` will be used instead.
        """
        return None

    @cached_property
    def index_template_name(self):
        return self.get_index_template()

    def get_index_template(self):
        """
        Returns a template to be used when rendering ``index_view``. If a
        template is specified by the ``index_template_name`` attribute, that will
        be used. Otherwise, a list of preferred template names are returned.

        **Deprecated** - the preferred way to customise this is to define an ``index_template_name`` property.
        """
        return self.get_templates("index")

    @cached_property
    def index_results_template_name(self):
        return self.get_index_results_template()

    def get_index_results_template(self):
        """
        Returns a template to be used when rendering ``index_results_view``. If a
        template is specified by the ``index_results_template_name`` attribute, that will
        be used. Otherwise, a list of preferred template names are returned.

        **Deprecated** - the preferred way to customise this is to define an ``index_results_template_name`` property.
        """
        return self.get_templates("index_results")

    @cached_property
    def create_template_name(self):
        return self.get_create_template()

    def get_create_template(self):
        """
        Returns a template to be used when rendering ``add_view``. If a
        template is specified by the ``create_template_name`` attribute, that will
        be used. Otherwise, a list of preferred template names are returned.

        **Deprecated** - the preferred way to customise this is to define a ``create_template_name`` property.
        """
        return self.get_templates("create")

    @cached_property
    def edit_template_name(self):
        return self.get_edit_template()

    def get_edit_template(self):
        """
        Returns a template to be used when rendering ``edit_view``. If a
        template is specified by the ``edit_template_name`` attribute, that will
        be used. Otherwise, a list of preferred template names are returned.

        **Deprecated** - the preferred way to customise this is to define an ``edit_template_name`` property.
        """
        return self.get_templates("edit")

    @cached_property
    def delete_template_name(self):
        return self.get_delete_template()

    def get_delete_template(self):
        """
        Returns a template to be used when rendering ``delete_view``. If a
        template is specified by the ``delete_template_name`` attribute, that will
        be used. Otherwise, a list of preferred template names are returned.

        **Deprecated** - the preferred way to customise this is to define a ``delete_template_name`` property.
        """
        return self.get_templates("delete")

    @cached_property
    def history_template_name(self):
        """
        A template to be used when rendering ``history_view``.

        Default: if :attr:`template_prefix` is specified, a ``history.html``
        template in the prefix directory and its ``{app_label}/{model_name}/``
        or ``{app_label}/`` subdirectories will be used. Otherwise, the
        ``history_view_class.template_name`` will be used.
        """
        return self.get_history_template()

    def get_history_template(self):
        """
        Returns a template to be used when rendering ``history_view``. If a
        template is specified by the ``history_template_name`` attribute, that will
        be used. Otherwise, a list of preferred template names are returned.

        **Deprecated** - the preferred way to customise this is to define a ``history_template_name`` property.
        """
        return self.get_templates("history")

    @cached_property
    def inspect_template_name(self):
        """
        A template to be used when rendering ``inspect_view``.

        Default: if :attr:`template_prefix` is specified, an ``inspect.html``
        template in the prefix directory and its ``{app_label}/{model_name}/``
        or ``{app_label}/`` subdirectories will be used. Otherwise, the
        ``inspect_view_class.template_name`` will be used.
        """
        return self.get_inspect_template()

    def get_inspect_template(self):
        """
        Returns a template to be used when rendering ``inspect_view``. If a
        template is specified by the ``inspect_template_name`` attribute, that will
        be used. Otherwise, a list of preferred template names are returned.

        **Deprecated** - the preferred way to customise this is to define an ``inspect_template_name`` property.
        """
        return self.get_templates(
            "inspect", fallback=self.inspect_view_class.template_name
        )

    def get_admin_url_namespace(self):
        """
        Returns the URL namespace for the admin URLs for this model.

        **Deprecated** - the preferred way to customise this is to define a ``url_namespace`` property.
        """
        if self.admin_url_namespace:
            return self.admin_url_namespace
        return f"wagtailsnippets_{self.app_label}_{self.model_name}"

    def get_admin_base_path(self):
        """
        Returns the base path for the admin URLs for this model.
        The returned string must not begin or end with a slash.

        **Deprecated** - the preferred way to customise this is to define a ``url_prefix`` property.
        """
        if self.base_url_path:
            return self.base_url_path.strip().strip("/")
        return f"snippets/{self.app_label}/{self.model_name}"

    def get_chooser_admin_url_namespace(self):
        """Returns the URL namespace for the chooser admin URLs for this model."""
        if self.chooser_admin_url_namespace:
            return self.chooser_admin_url_namespace
        return f"wagtailsnippetchoosers_{self.app_label}_{self.model_name}"

    def get_chooser_admin_base_path(self):
        """
        Returns the base path for the chooser admin URLs for this model.
        The returned string must not begin or end with a slash.
        """
        if self.chooser_base_url_path:
            return self.chooser_base_url_path.strip().strip("/")
        return f"snippets/choose/{self.app_label}/{self.model_name}"

    @property
    def url_finder_class(self):
        return type(
            "_SnippetAdminURLFinder", (SnippetAdminURLFinder,), {"model": self.model}
        )

    def get_urlpatterns(self):
        urlpatterns = [
            path("", self.index_view, name="list"),
            path("results/", self.index_results_view, name="list_results"),
            path("add/", self.add_view, name="add"),
            path("edit/<str:pk>/", self.edit_view, name="edit"),
            path("delete/<str:pk>/", self.delete_view, name="delete"),
            path("usage/<str:pk>/", self.usage_view, name="usage"),
            path("history/<str:pk>/", self.history_view, name="history"),
            path(
                "history-results/<str:pk>/",
                self.history_results_view,
                name="history_results",
            ),
        ]

        if self.copy_view_enabled:
            urlpatterns += [path("copy/<str:pk>/", self.copy_view, name="copy")]

        if self.inspect_view_enabled:
            urlpatterns += [
                path("inspect/<str:pk>/", self.inspect_view, name="inspect")
            ]

        if self.preview_enabled:
            urlpatterns += [
                path("preview/", self.preview_on_add_view, name="preview_on_add"),
                path(
                    "preview/<str:pk>/",
                    self.preview_on_edit_view,
                    name="preview_on_edit",
                ),
            ]

        if self.revision_enabled:
            if self.preview_enabled:
                urlpatterns += [
                    path(
                        "history/<str:pk>/revisions/<int:revision_id>/view/",
                        self.revisions_view,
                        name="revisions_view",
                    )
                ]

            urlpatterns += [
                path(
                    "history/<str:pk>/revisions/<int:revision_id>/revert/",
                    self.revisions_revert_view,
                    name="revisions_revert",
                ),
                re_path(
                    r"history/(?P<pk>.+)/revisions/compare/(?P<revision_id_a>live|earliest|\d+)\.\.\.(?P<revision_id_b>live|latest|\d+)/$",
                    self.revisions_compare_view,
                    name="revisions_compare",
                ),
            ]

        if self.draftstate_enabled:
            urlpatterns += [
                path(
                    "history/<str:pk>/revisions/<int:revision_id>/unschedule/",
                    self.revisions_unschedule_view,
                    name="revisions_unschedule",
                ),
                path("unpublish/<str:pk>/", self.unpublish_view, name="unpublish"),
            ]

        if self.locking_enabled:
            urlpatterns += [
                path("lock/<str:pk>/", self.lock_view, name="lock"),
                path("unlock/<str:pk>/", self.unlock_view, name="unlock"),
            ]

        if self.workflow_enabled:
            urlpatterns += [
                path(
                    "workflow/action/<str:pk>/<slug:action_name>/<int:task_state_id>/",
                    self.workflow_action_view,
                    name="workflow_action",
                ),
                path(
                    "workflow/collect_action_data/<str:pk>/<slug:action_name>/<int:task_state_id>/",
                    self.collect_workflow_action_data_view,
                    name="collect_workflow_action_data",
                ),
                path(
                    "workflow/confirm_cancellation/<str:pk>/",
                    self.confirm_workflow_cancellation_view,
                    name="confirm_workflow_cancellation",
                ),
                path(
                    "workflow_history/<str:pk>/",
                    self.workflow_history_view,
                    name="workflow_history",
                ),
                path(
                    "workflow_history/<str:pk>/detail/<int:workflow_state_id>/",
                    self.workflow_history_detail_view,
                    name="workflow_history_detail",
                ),
            ]

            if self.preview_enabled:
                urlpatterns += [
                    path(
                        "workflow/preview/<str:pk>/<int:task_id>/",
                        self.workflow_preview_view,
                        name="workflow_preview",
                    ),
                ]

        return urlpatterns

    def get_edit_handler(self):
        """
        Like :meth:`ModelViewSet.get_edit_handler()
        <wagtail.admin.viewsets.model.ModelViewSet.get_edit_handler>`,
        but falls back to extracting panel definitions from the model class
        if no edit handler is defined.
        """
        edit_handler = super().get_edit_handler()
        if edit_handler:
            return edit_handler

        exclude = self.get_exclude_form_fields()
        panels = extract_panel_definitions_from_model_class(self.model, exclude=exclude)
        edit_handler = ObjectList(panels)
        return edit_handler.bind_to_model(self.model)

    def register_chooser_viewset(self):
        viewsets.register(self.chooser_viewset)

    def register_model_check(self):
        def snippets_model_check(app_configs, **kwargs):
            return check_panels_in_model(self.model, "snippets")

        checks.register(snippets_model_check, "panels")

    def register_snippet_model(self):
        snippet_models = get_snippet_models()
        if self.model in snippet_models:
            raise ImproperlyConfigured(
                f"The {self.model.__name__} model is already registered as a snippet"
            )
        snippet_models.append(self.model)
        snippet_models.sort(key=lambda x: x._meta.verbose_name)

    def on_register(self):
        super().on_register()
        # For convenience, attach viewset to the model class to allow accessing
        # the configuration of a given model.
        self.model.snippet_viewset = self
        self.register_chooser_viewset()
        self.register_model_check()
        self.register_snippet_model()


class SnippetViewSetGroup(ModelViewSetGroup):
    """
    A container for grouping together multiple
    :class:`~wagtail.snippets.views.snippets.SnippetViewSet` instances.

    All attributes and methods from
    :class:`~wagtail.admin.viewsets.model.ModelViewSetGroup` are available.
    """

    def __init__(self):
        menu_item_is_registered = getattr(
            self, "menu_item_is_registered", bool(self.menu_hook)
        )
        # If the menu item is registered, mark all viewsets as such so that we can
        # hide the "Snippets" menu item if all snippets have their own menu items.
        for item in self.items:
            item.menu_item_is_registered = menu_item_is_registered

        # Call super() after setting menu_item_is_registered so that nested groups
        # can inherit the value.
        super().__init__()



import django.db.models.deletion
import modelcluster.contrib.taggit
import modelcluster.fields
from django.db import migrations, models

import wagtail.fields


class Migration(migrations.Migration):

    dependencies = [
        ("wagtailimages", "0005_make_filter_spec_unique"),
        ("taggit", "0001_initial"),
        ("wagtaildocs", "0002_initial_data"),
        ("wagtailcore", "0013_update_golive_expire_help_text"),
    ]

    operations = [
        migrations.CreateModel(
            name="BlogEntryPage",
            fields=[
                (
                    "page_ptr",
                    models.OneToOneField(
                        on_delete=models.CASCADE,
                        to="wagtailcore.Page",
                        serialize=False,
                        parent_link=True,
                        related_name="+",
                        primary_key=True,
                    ),
                ),
                ("body", wagtail.fields.RichTextField()),
                ("date", models.DateField(verbose_name="Post date")),
                (
                    "feed_image",
                    models.ForeignKey(
                        to="wagtailimages.Image",
                        blank=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        null=True,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("wagtailcore.page",),
        ),
        migrations.CreateModel(
            name="BlogEntryPageCarouselItem",
            fields=[
                (
                    "id",
                    models.AutoField(
                        serialize=False,
                        verbose_name="ID",
                        primary_key=True,
                        auto_created=True,
                    ),
                ),
                (
                    "sort_order",
                    models.IntegerField(null=True, editable=False, blank=True),
                ),
                (
                    "link_external",
                    models.URLField(blank=True, verbose_name="External link"),
                ),
                ("embed_url", models.URLField(blank=True, verbose_name="Embed URL")),
                ("caption", models.CharField(blank=True, max_length=255)),
                (
                    "image",
                    models.ForeignKey(
                        to="wagtailimages.Image",
                        blank=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        null=True,
                    ),
                ),
                (
                    "link_document",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        to="wagtaildocs.Document",
                        blank=True,
                        related_name="+",
                        null=True,
                    ),
                ),
            ],
            options={
                "ordering": ["sort_order"],
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="BlogEntryPageRelatedLink",
            fields=[
                (
                    "id",
                    models.AutoField(
                        serialize=False,
                        verbose_name="ID",
                        primary_key=True,
                        auto_created=True,
                    ),
                ),
                (
                    "sort_order",
                    models.IntegerField(null=True, editable=False, blank=True),
                ),
                (
                    "link_external",
                    models.URLField(blank=True, verbose_name="External link"),
                ),
                ("title", models.CharField(help_text="Link title", max_length=255)),
                (
                    "link_document",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        to="wagtaildocs.Document",
                        blank=True,
                        related_name="+",
                        null=True,
                    ),
                ),
            ],
            options={
                "ordering": ["sort_order"],
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="BlogEntryPageTag",
            fields=[
                (
                    "id",
                    models.AutoField(
                        serialize=False,
                        verbose_name="ID",
                        primary_key=True,
                        auto_created=True,
                    ),
                ),
                (
                    "content_object",
                    modelcluster.fields.ParentalKey(
                        on_delete=models.CASCADE,
                        to="demosite.BlogEntryPage",
                        related_name="tagged_items",
                    ),
                ),
                (
                    "tag",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        to="taggit.Tag",
                        related_name="demosite_blogentrypagetag_items",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="BlogIndexPage",
            fields=[
                (
                    "page_ptr",
                    models.OneToOneField(
                        on_delete=models.CASCADE,
                        to="wagtailcore.Page",
                        serialize=False,
                        parent_link=True,
                        related_name="+",
                        primary_key=True,
                    ),
                ),
                ("intro", wagtail.fields.RichTextField(blank=True)),
            ],
            options={
                "abstract": False,
            },
            bases=("wagtailcore.page",),
        ),
        migrations.CreateModel(
            name="BlogIndexPageRelatedLink",
            fields=[
                (
                    "id",
                    models.AutoField(
                        serialize=False,
                        verbose_name="ID",
                        primary_key=True,
                        auto_created=True,
                    ),
                ),
                (
                    "sort_order",
                    models.IntegerField(null=True, editable=False, blank=True),
                ),
                (
                    "link_external",
                    models.URLField(blank=True, verbose_name="External link"),
                ),
                ("title", models.CharField(help_text="Link title", max_length=255)),
                (
                    "link_document",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        to="wagtaildocs.Document",
                        blank=True,
                        related_name="+",
                        null=True,
                    ),
                ),
            ],
            options={
                "ordering": ["sort_order"],
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="ContactPage",
            fields=[
                ("telephone", models.CharField(blank=True, max_length=20)),
                ("email", models.EmailField(blank=True, max_length=254)),
                ("address_1", models.CharField(blank=True, max_length=255)),
                ("address_2", models.CharField(blank=True, max_length=255)),
                ("city", models.CharField(blank=True, max_length=255)),
                ("country", models.CharField(blank=True, max_length=255)),
                ("post_code", models.CharField(blank=True, max_length=10)),
                (
                    "page_ptr",
                    models.OneToOneField(
                        on_delete=models.CASCADE,
                        to="wagtailcore.Page",
                        serialize=False,
                        parent_link=True,
                        related_name="+",
                        primary_key=True,
                    ),
                ),
                ("body", wagtail.fields.RichTextField(blank=True)),
                (
                    "feed_image",
                    models.ForeignKey(
                        to="wagtailimages.Image",
                        blank=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        null=True,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("wagtailcore.page", models.Model),
        ),
        migrations.CreateModel(
            name="EventIndexPage",
            fields=[
                (
                    "page_ptr",
                    models.OneToOneField(
                        on_delete=models.CASCADE,
                        to="wagtailcore.Page",
                        serialize=False,
                        parent_link=True,
                        related_name="+",
                        primary_key=True,
                    ),
                ),
                ("intro", wagtail.fields.RichTextField(blank=True)),
            ],
            options={
                "abstract": False,
            },
            bases=("wagtailcore.page",),
        ),
        migrations.CreateModel(
            name="EventIndexPageRelatedLink",
            fields=[
                (
                    "id",
                    models.AutoField(
                        serialize=False,
                        verbose_name="ID",
                        primary_key=True,
                        auto_created=True,
                    ),
                ),
                (
                    "sort_order",
                    models.IntegerField(null=True, editable=False, blank=True),
                ),
                (
                    "link_external",
                    models.URLField(blank=True, verbose_name="External link"),
                ),
                ("title", models.CharField(help_text="Link title", max_length=255)),
                (
                    "link_document",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        to="wagtaildocs.Document",
                        blank=True,
                        related_name="+",
                        null=True,
                    ),
                ),
            ],
            options={
                "ordering": ["sort_order"],
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="EventPage",
            fields=[
                (
                    "page_ptr",
                    models.OneToOneField(
                        on_delete=models.CASCADE,
                        to="wagtailcore.Page",
                        serialize=False,
                        parent_link=True,
                        related_name="+",
                        primary_key=True,
                    ),
                ),
                ("date_from", models.DateField(verbose_name="Start date")),
                (
                    "date_to",
                    models.DateField(
                        help_text="Not required if event is on a single day",
                        null=True,
                        verbose_name="End date",
                        blank=True,
                    ),
                ),
                (
                    "time_from",
                    models.TimeField(null=True, verbose_name="Start time", blank=True),
                ),
                (
                    "time_to",
                    models.TimeField(null=True, verbose_name="End time", blank=True),
                ),
                (
                    "audience",
                    models.CharField(
                        choices=[("public", "Public"), ("private", "Private")],
                        max_length=255,
                    ),
                ),
                ("location", models.CharField(max_length=255)),
                ("body", wagtail.fields.RichTextField(blank=True)),
                ("cost", models.CharField(max_length=255)),
                ("signup_link", models.URLField(blank=True)),
                (
                    "feed_image",
                    models.ForeignKey(
                        to="wagtailimages.Image",
                        blank=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        null=True,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("wagtailcore.page",),
        ),
        migrations.CreateModel(
            name="EventPageCarouselItem",
            fields=[
                (
                    "id",
                    models.AutoField(
                        serialize=False,
                        verbose_name="ID",
                        primary_key=True,
                        auto_created=True,
                    ),
                ),
                (
                    "sort_order",
                    models.IntegerField(null=True, editable=False, blank=True),
                ),
                (
                    "link_external",
                    models.URLField(blank=True, verbose_name="External link"),
                ),
                ("embed_url", models.URLField(blank=True, verbose_name="Embed URL")),
                ("caption", models.CharField(blank=True, max_length=255)),
                (
                    "image",
                    models.ForeignKey(
                        to="wagtailimages.Image",
                        blank=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        null=True,
                    ),
                ),
                (
                    "link_document",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        to="wagtaildocs.Document",
                        blank=True,
                        related_name="+",
                        null=True,
                    ),
                ),
            ],
            options={
                "ordering": ["sort_order"],
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="EventPageRelatedLink",
            fields=[
                (
                    "id",
                    models.AutoField(
                        serialize=False,
                        verbose_name="ID",
                        primary_key=True,
                        auto_created=True,
                    ),
                ),
                (
                    "sort_order",
                    models.IntegerField(null=True, editable=False, blank=True),
                ),
                (
                    "link_external",
                    models.URLField(blank=True, verbose_name="External link"),
                ),
                ("title", models.CharField(help_text="Link title", max_length=255)),
                (
                    "link_document",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        to="wagtaildocs.Document",
                        blank=True,
                        related_name="+",
                        null=True,
                    ),
                ),
            ],
            options={
                "ordering": ["sort_order"],
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="EventPageSpeaker",
            fields=[
                (
                    "id",
                    models.AutoField(
                        serialize=False,
                        verbose_name="ID",
                        primary_key=True,
                        auto_created=True,
                    ),
                ),
                (
                    "sort_order",
                    models.IntegerField(null=True, editable=False, blank=True),
                ),
                (
                    "link_external",
                    models.URLField(blank=True, verbose_name="External link"),
                ),
                (
                    "first_name",
                    models.CharField(blank=True, verbose_name="Name", max_length=255),
                ),
                (
                    "last_name",
                    models.CharField(
                        blank=True, verbose_name="Surname", max_length=255
                    ),
                ),
                (
                    "image",
                    models.ForeignKey(
                        to="wagtailimages.Image",
                        blank=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        null=True,
                    ),
                ),
                (
                    "link_document",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        to="wagtaildocs.Document",
                        blank=True,
                        related_name="+",
                        null=True,
                    ),
                ),
            ],
            options={
                "ordering": ["sort_order"],
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="HomePage",
            fields=[
                (
                    "page_ptr",
                    models.OneToOneField(
                        on_delete=models.CASCADE,
                        to="wagtailcore.Page",
                        serialize=False,
                        parent_link=True,
                        related_name="+",
                        primary_key=True,
                    ),
                ),
                ("body", wagtail.fields.RichTextField(blank=True)),
            ],
            options={
                "verbose_name": "Homepage",
            },
            bases=("wagtailcore.page",),
        ),
        migrations.CreateModel(
            name="HomePageCarouselItem",
            fields=[
                (
                    "id",
                    models.AutoField(
                        serialize=False,
                        verbose_name="ID",
                        primary_key=True,
                        auto_created=True,
                    ),
                ),
                (
                    "sort_order",
                    models.IntegerField(null=True, editable=False, blank=True),
                ),
                (
                    "link_external",
                    models.URLField(blank=True, verbose_name="External link"),
                ),
                ("embed_url", models.URLField(blank=True, verbose_name="Embed URL")),
                ("caption", models.CharField(blank=True, max_length=255)),
                (
                    "image",
                    models.ForeignKey(
                        to="wagtailimages.Image",
                        blank=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        null=True,
                    ),
                ),
                (
                    "link_document",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        to="wagtaildocs.Document",
                        blank=True,
                        related_name="+",
                        null=True,
                    ),
                ),
            ],
            options={
                "ordering": ["sort_order"],
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="HomePageRelatedLink",
            fields=[
                (
                    "id",
                    models.AutoField(
                        serialize=False,
                        verbose_name="ID",
                        primary_key=True,
                        auto_created=True,
                    ),
                ),
                (
                    "sort_order",
                    models.IntegerField(null=True, editable=False, blank=True),
                ),
                (
                    "link_external",
                    models.URLField(blank=True, verbose_name="External link"),
                ),
                ("title", models.CharField(help_text="Link title", max_length=255)),
                (
                    "link_document",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        to="wagtaildocs.Document",
                        blank=True,
                        related_name="+",
                        null=True,
                    ),
                ),
            ],
            options={
                "ordering": ["sort_order"],
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="PersonPage",
            fields=[
                ("telephone", models.CharField(blank=True, max_length=20)),
                ("email", models.EmailField(blank=True, max_length=254)),
                ("address_1", models.CharField(blank=True, max_length=255)),
                ("address_2", models.CharField(blank=True, max_length=255)),
                ("city", models.CharField(blank=True, max_length=255)),
                ("country", models.CharField(blank=True, max_length=255)),
                ("post_code", models.CharField(blank=True, max_length=10)),
                (
                    "page_ptr",
                    models.OneToOneField(
                        on_delete=models.CASCADE,
                        to="wagtailcore.Page",
                        serialize=False,
                        parent_link=True,
                        related_name="+",
                        primary_key=True,
                    ),
                ),
                ("first_name", models.CharField(max_length=255)),
                ("last_name", models.CharField(max_length=255)),
                ("intro", wagtail.fields.RichTextField(blank=True)),
                ("biography", wagtail.fields.RichTextField(blank=True)),
                (
                    "feed_image",
                    models.ForeignKey(
                        to="wagtailimages.Image",
                        blank=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        null=True,
                    ),
                ),
                (
                    "image",
                    models.ForeignKey(
                        to="wagtailimages.Image",
                        blank=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        null=True,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("wagtailcore.page", models.Model),
        ),
        migrations.CreateModel(
            name="PersonPageRelatedLink",
            fields=[
                (
                    "id",
                    models.AutoField(
                        serialize=False,
                        verbose_name="ID",
                        primary_key=True,
                        auto_created=True,
                    ),
                ),
                (
                    "sort_order",
                    models.IntegerField(null=True, editable=False, blank=True),
                ),
                (
                    "link_external",
                    models.URLField(blank=True, verbose_name="External link"),
                ),
                ("title", models.CharField(help_text="Link title", max_length=255)),
                (
                    "link_document",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        to="wagtaildocs.Document",
                        blank=True,
                        related_name="+",
                        null=True,
                    ),
                ),
            ],
            options={
                "ordering": ["sort_order"],
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="StandardIndexPage",
            fields=[
                (
                    "page_ptr",
                    models.OneToOneField(
                        on_delete=models.CASCADE,
                        to="wagtailcore.Page",
                        serialize=False,
                        parent_link=True,
                        related_name="+",
                        primary_key=True,
                    ),
                ),
                ("intro", wagtail.fields.RichTextField(blank=True)),
                (
                    "feed_image",
                    models.ForeignKey(
                        to="wagtailimages.Image",
                        blank=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        null=True,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("wagtailcore.page",),
        ),
        migrations.CreateModel(
            name="StandardIndexPageRelatedLink",
            fields=[
                (
                    "id",
                    models.AutoField(
                        serialize=False,
                        verbose_name="ID",
                        primary_key=True,
                        auto_created=True,
                    ),
                ),
                (
                    "sort_order",
                    models.IntegerField(null=True, editable=False, blank=True),
                ),
                (
                    "link_external",
                    models.URLField(blank=True, verbose_name="External link"),
                ),
                ("title", models.CharField(help_text="Link title", max_length=255)),
                (
                    "link_document",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        to="wagtaildocs.Document",
                        blank=True,
                        related_name="+",
                        null=True,
                    ),
                ),
            ],
            options={
                "ordering": ["sort_order"],
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="StandardPage",
            fields=[
                (
                    "page_ptr",
                    models.OneToOneField(
                        on_delete=models.CASCADE,
                        to="wagtailcore.Page",
                        serialize=False,
                        parent_link=True,
                        related_name="+",
                        primary_key=True,
                    ),
                ),
                ("intro", wagtail.fields.RichTextField(blank=True)),
                ("body", wagtail.fields.RichTextField(blank=True)),
                (
                    "feed_image",
                    models.ForeignKey(
                        to="wagtailimages.Image",
                        blank=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        null=True,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("wagtailcore.page",),
        ),
        migrations.CreateModel(
            name="StandardPageCarouselItem",
            fields=[
                (
                    "id",
                    models.AutoField(
                        serialize=False,
                        verbose_name="ID",
                        primary_key=True,
                        auto_created=True,
                    ),
                ),
                (
                    "sort_order",
                    models.IntegerField(null=True, editable=False, blank=True),
                ),
                (
                    "link_external",
                    models.URLField(blank=True, verbose_name="External link"),
                ),
                ("embed_url", models.URLField(blank=True, verbose_name="Embed URL")),
                ("caption", models.CharField(blank=True, max_length=255)),
                (
                    "image",
                    models.ForeignKey(
                        to="wagtailimages.Image",
                        blank=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        null=True,
                    ),
                ),
                (
                    "link_document",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        to="wagtaildocs.Document",
                        blank=True,
                        related_name="+",
                        null=True,
                    ),
                ),
                (
                    "link_page",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        to="wagtailcore.Page",
                        blank=True,
                        related_name="+",
                        null=True,
                    ),
                ),
                (
                    "page",
                    modelcluster.fields.ParentalKey(
                        on_delete=models.CASCADE,
                        to="demosite.StandardPage",
                        related_name="carousel_items",
                    ),
                ),
            ],
            options={
                "ordering": ["sort_order"],
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.CreateModel(
            name="StandardPageRelatedLink",
            fields=[
                (
                    "id",
                    models.AutoField(
                        serialize=False,
                        verbose_name="ID",
                        primary_key=True,
                        auto_created=True,
                    ),
                ),
                (
                    "sort_order",
                    models.IntegerField(null=True, editable=False, blank=True),
                ),
                (
                    "link_external",
                    models.URLField(blank=True, verbose_name="External link"),
                ),
                ("title", models.CharField(help_text="Link title", max_length=255)),
                (
                    "link_document",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        to="wagtaildocs.Document",
                        blank=True,
                        related_name="+",
                        null=True,
                    ),
                ),
                (
                    "link_page",
                    models.ForeignKey(
                        on_delete=models.CASCADE,
                        to="wagtailcore.Page",
                        blank=True,
                        related_name="+",
                        null=True,
                    ),
                ),
                (
                    "page",
                    modelcluster.fields.ParentalKey(
                        on_delete=models.CASCADE,
                        to="demosite.StandardPage",
                        related_name="related_links",
                    ),
                ),
            ],
            options={
                "ordering": ["sort_order"],
                "abstract": False,
            },
            bases=(models.Model,),
        ),
        migrations.AddField(
            model_name="standardindexpagerelatedlink",
            name="link_page",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                to="wagtailcore.Page",
                blank=True,
                related_name="+",
                null=True,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="standardindexpagerelatedlink",
            name="page",
            field=modelcluster.fields.ParentalKey(
                on_delete=models.CASCADE,
                to="demosite.StandardIndexPage",
                related_name="related_links",
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="personpagerelatedlink",
            name="link_page",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                to="wagtailcore.Page",
                blank=True,
                related_name="+",
                null=True,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="personpagerelatedlink",
            name="page",
            field=modelcluster.fields.ParentalKey(
                on_delete=models.CASCADE,
                to="demosite.PersonPage",
                related_name="related_links",
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="homepagerelatedlink",
            name="link_page",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                to="wagtailcore.Page",
                blank=True,
                related_name="+",
                null=True,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="homepagerelatedlink",
            name="page",
            field=modelcluster.fields.ParentalKey(
                on_delete=models.CASCADE,
                to="demosite.HomePage",
                related_name="related_links",
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="homepagecarouselitem",
            name="link_page",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                to="wagtailcore.Page",
                blank=True,
                related_name="+",
                null=True,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="homepagecarouselitem",
            name="page",
            field=modelcluster.fields.ParentalKey(
                on_delete=models.CASCADE,
                to="demosite.HomePage",
                related_name="carousel_items",
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="eventpagespeaker",
            name="link_page",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                to="wagtailcore.Page",
                blank=True,
                related_name="+",
                null=True,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="eventpagespeaker",
            name="page",
            field=modelcluster.fields.ParentalKey(
                on_delete=models.CASCADE,
                to="demosite.EventPage",
                related_name="speakers",
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="eventpagerelatedlink",
            name="link_page",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                to="wagtailcore.Page",
                blank=True,
                related_name="+",
                null=True,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="eventpagerelatedlink",
            name="page",
            field=modelcluster.fields.ParentalKey(
                on_delete=models.CASCADE,
                to="demosite.EventPage",
                related_name="related_links",
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="eventpagecarouselitem",
            name="link_page",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                to="wagtailcore.Page",
                blank=True,
                related_name="+",
                null=True,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="eventpagecarouselitem",
            name="page",
            field=modelcluster.fields.ParentalKey(
                on_delete=models.CASCADE,
                to="demosite.EventPage",
                related_name="carousel_items",
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="eventindexpagerelatedlink",
            name="link_page",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                to="wagtailcore.Page",
                blank=True,
                related_name="+",
                null=True,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="eventindexpagerelatedlink",
            name="page",
            field=modelcluster.fields.ParentalKey(
                on_delete=models.CASCADE,
                to="demosite.EventIndexPage",
                related_name="related_links",
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="blogindexpagerelatedlink",
            name="link_page",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                to="wagtailcore.Page",
                blank=True,
                related_name="+",
                null=True,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="blogindexpagerelatedlink",
            name="page",
            field=modelcluster.fields.ParentalKey(
                on_delete=models.CASCADE,
                to="demosite.BlogIndexPage",
                related_name="related_links",
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="blogentrypagerelatedlink",
            name="link_page",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                to="wagtailcore.Page",
                blank=True,
                related_name="+",
                null=True,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="blogentrypagerelatedlink",
            name="page",
            field=modelcluster.fields.ParentalKey(
                on_delete=models.CASCADE,
                to="demosite.BlogEntryPage",
                related_name="related_links",
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="blogentrypagecarouselitem",
            name="link_page",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                to="wagtailcore.Page",
                blank=True,
                related_name="+",
                null=True,
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="blogentrypagecarouselitem",
            name="page",
            field=modelcluster.fields.ParentalKey(
                on_delete=models.CASCADE,
                to="demosite.BlogEntryPage",
                related_name="carousel_items",
            ),
            preserve_default=True,
        ),
        migrations.AddField(
            model_name="blogentrypage",
            name="tags",
            field=modelcluster.contrib.taggit.ClusterTaggableManager(
                help_text="A comma-separated list of tags.",
                through="demosite.BlogEntryPageTag",
                blank=True,
                verbose_name="Tags",
                to="taggit.Tag",
            ),
            preserve_default=True,
        ),
        migrations.CreateModel(
            name="FormPage",
            fields=[
                (
                    "page_ptr",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="+",
                        serialize=False,
                        to="wagtailcore.page",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("wagtailcore.page",),
        ),
        migrations.CreateModel(
            name="FormField",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "sort_order",
                    models.IntegerField(blank=True, editable=False, null=True),
                ),
                (
                    "clean_name",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Safe name of the form field, the label converted to ascii_snake_case",
                        max_length=255,
                        verbose_name="name",
                    ),
                ),
                (
                    "label",
                    models.CharField(
                        help_text="The label of the form field",
                        max_length=255,
                        verbose_name="label",
                    ),
                ),
                (
                    "field_type",
                    models.CharField(
                        choices=[
                            ("singleline", "Single line text"),
                            ("multiline", "Multi-line text"),
                            ("email", "Email"),
                            ("number", "Number"),
                            ("url", "URL"),
                            ("checkbox", "Checkbox"),
                            ("checkboxes", "Checkboxes"),
                            ("dropdown", "Drop down"),
                            ("multiselect", "Multiple select"),
                            ("radio", "Radio buttons"),
                            ("date", "Date"),
                            ("datetime", "Date/time"),
                            ("hidden", "Hidden field"),
                        ],
                        max_length=16,
                        verbose_name="field type",
                    ),
                ),
                (
                    "required",
                    models.BooleanField(default=True, verbose_name="required"),
                ),
                (
                    "choices",
                    models.TextField(
                        blank=True,
                        help_text="Comma or new line separated list of choices. Only applicable in checkboxes, radio and dropdown.",
                        verbose_name="choices",
                    ),
                ),
                (
                    "default_value",
                    models.TextField(
                        blank=True,
                        help_text="Default value. Comma or new line separated values supported for checkboxes.",
                        verbose_name="default value",
                    ),
                ),
                (
                    "help_text",
                    models.CharField(
                        blank=True, max_length=255, verbose_name="help text"
                    ),
                ),
                (
                    "page",
                    modelcluster.fields.ParentalKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="form_fields",
                        to="demosite.formpage",
                    ),
                ),
            ],
            options={
                "ordering": ["sort_order"],
                "abstract": False,
            },
        ),
    ]



import os

from django.core.files.uploadedfile import SimpleUploadedFile
from django.test import TestCase, override_settings
from django.urls import reverse

from wagtail.contrib.redirects.models import Redirect
from wagtail.models import Site
from wagtail.test.utils import WagtailTestUtils

TEST_ROOT = os.path.abspath(os.path.dirname(__file__))


@override_settings(
    ALLOWED_HOSTS=["testserver", "localhost", "test.example.com", "other.example.com"]
)
class TestImportAdminViews(WagtailTestUtils, TestCase):
    def setUp(self):
        self.login()

    def get(self, params={}):
        return self.client.get(reverse("wagtailredirects:start_import"), params)

    def post(self, post_data={}, follow=False):
        return self.client.post(
            reverse("wagtailredirects:start_import"), post_data, follow=follow
        )

    def post_import(self, post_data={}, follow=False):
        return self.client.post(
            reverse("wagtailredirects:process_import"), post_data, follow=follow
        )

    def test_request_start_with_get_returns_initial_form(self):
        response = self.get()
        self.assertEqual(
            response.templates[0].name,
            "wagtailredirects/choose_import_file.html",
        )

    def test_empty_import_file_returns_error(self):
        response = self.post(
            {
                "import_file": "",
                "input_format": "0",
            }
        )

        self.assertIn("import_file", response.context["form"].errors)

    def test_non_valid_format_returns_error(self):
        f = f"{TEST_ROOT}/files/example.yaml"
        (_, filename) = os.path.split(f)

        with open(f, "rb") as infile:
            upload_file = SimpleUploadedFile(filename, infile.read())

            response = self.post(
                {
                    "import_file": upload_file,
                },
                follow=True,
            )

            self.assertContains(
                response, "File format of type &quot;yaml&quot; is not supported"
            )

    def test_valid_csv_triggers_confirm_view(self):
        f = f"{TEST_ROOT}/files/example.csv"
        (_, filename) = os.path.split(f)

        with open(f, "rb") as infile:
            upload_file = SimpleUploadedFile(filename, infile.read())

            response = self.post(
                {
                    "import_file": upload_file,
                }
            )

            self.assertEqual(
                response.templates[0].name,
                "wagtailredirects/confirm_import.html",
            )
            self.assertEqual(len(response.context["dataset"]), 3)

    def test_import_step(self):
        f = f"{TEST_ROOT}/files/example.csv"
        (_, filename) = os.path.split(f)

        with open(f, "rb") as infile:
            upload_file = SimpleUploadedFile(filename, infile.read())

            self.assertEqual(Redirect.objects.all().count(), 0)

            response = self.post(
                {
                    "import_file": upload_file,
                }
            )

            import_response = self.post_import(
                {
                    **response.context["form"].initial,
                    "from_index": 0,
                    "to_index": 1,
                    "permanent": True,
                }
            )

            self.assertTemplateUsed(
                import_response, "wagtailredirects/import_summary.html"
            )
            self.assertEqual(Redirect.objects.all().count(), 2)

    def test_import_step_with_offset_columns(self):
        f = f"{TEST_ROOT}/files/example_offset_columns.csv"
        (_, filename) = os.path.split(f)

        with open(f, "rb") as infile:
            upload_file = SimpleUploadedFile(filename, infile.read())

            self.assertEqual(Redirect.objects.all().count(), 0)

            response = self.post(
                {
                    "import_file": upload_file,
                }
            )

            import_response = self.post_import(
                {
                    **response.context["form"].initial,
                    "from_index": 16,
                    "to_index": 17,
                    "permanent": True,
                },
            )

            self.assertTemplateUsed(
                import_response, "wagtailredirects/import_summary.html"
            )
            self.assertEqual(Redirect.objects.all().count(), 2)

    def test_permanent_setting(self):
        f = f"{TEST_ROOT}/files/example.csv"
        (_, filename) = os.path.split(f)

        with open(f, "rb") as infile:
            upload_file = SimpleUploadedFile(filename, infile.read())

            self.assertEqual(Redirect.objects.all().count(), 0)

            response = self.post(
                {
                    "import_file": upload_file,
                }
            )

            import_response = self.post_import(
                {
                    **response.context["form"].initial,
                    "from_index": 0,
                    "to_index": 1,
                    "permanent": False,
                }
            )

            self.assertTemplateUsed(
                import_response, "wagtailredirects/import_summary.html"
            )
            self.assertFalse(Redirect.objects.first().is_permanent)

    def test_site_setting(self):
        f = f"{TEST_ROOT}/files/example.csv"
        (_, filename) = os.path.split(f)

        default_site = Site.objects.first()
        new_site = Site.objects.create(
            hostname="hello.dev",
            root_page=default_site.root_page,
        )

        with open(f, "rb") as infile:
            upload_file = SimpleUploadedFile(filename, infile.read())

            self.assertEqual(Redirect.objects.all().count(), 0)

            response = self.post(
                {
                    "import_file": upload_file,
                }
            )

            import_response = self.post_import(
                {
                    **response.context["form"].initial,
                    "from_index": 0,
                    "to_index": 1,
                    "permanent": False,
                    "site": new_site.pk,
                }
            )

            self.assertTemplateUsed(
                import_response, "wagtailredirects/import_summary.html"
            )
            self.assertEqual(Redirect.objects.count(), 2)
            self.assertEqual(Redirect.objects.first().site, new_site)

    def test_import_xlsx(self):
        f = f"{TEST_ROOT}/files/example.xlsx"
        (_, filename) = os.path.split(f)

        with open(f, "rb") as infile:
            upload_file = SimpleUploadedFile(filename, infile.read())

            self.assertEqual(Redirect.objects.all().count(), 0)

            response = self.post(
                {
                    "import_file": upload_file,
                }
            )

            self.assertTemplateUsed(response, "wagtailredirects/confirm_import.html")

            import_response = self.post_import(
                {
                    **response.context["form"].initial,
                    "from_index": 0,
                    "to_index": 1,
                    "permanent": True,
                },
                follow=True,
            )

            self.assertTemplateUsed(import_response, "wagtailredirects/index.html")
            self.assertEqual(Redirect.objects.all().count(), 3)

    def test_unicode_error_when_importing(self):
        f = f"{TEST_ROOT}/files/example_faulty.csv"
        (_, filename) = os.path.split(f)

        with open(f, "rb") as infile:
            upload_file = SimpleUploadedFile(filename, infile.read())

            self.assertEqual(Redirect.objects.all().count(), 0)

            response = self.post(
                {
                    "import_file": upload_file,
                },
                follow=True,
            )
            self.assertIn(b"Imported file has a wrong encoding:", response.content)

    def test_not_valid_method_for_import_file(self):
        response = self.client.get(reverse("wagtailredirects:process_import"))
        self.assertEqual(response.status_code, 405)

    def test_error_in_data_renders_confirm_view_on_import(self):
        f = f"{TEST_ROOT}/files/example.csv"
        (_, filename) = os.path.split(f)

        with open(f, "rb") as infile:
            upload_file = SimpleUploadedFile(filename, infile.read())

            response = self.post(
                {
                    "import_file": upload_file,
                }
            )

            self.post_import(
                {
                    **response.context["form"].initial,
                    "from_index": 0,
                    "to_index": 1,
                    "permanent": True,
                    "site": 99,
                }
            )
            self.assertTemplateUsed(response, "wagtailredirects/confirm_import.html")

    def test_import_tsv(self):
        f = f"{TEST_ROOT}/files/example.tsv"
        (_, filename) = os.path.split(f)

        with open(f, "rb") as infile:
            upload_file = SimpleUploadedFile(filename, infile.read())

            self.assertEqual(Redirect.objects.all().count(), 0)

            response = self.post(
                {
                    "import_file": upload_file,
                }
            )

            self.assertTemplateUsed(response, "wagtailredirects/confirm_import.html")

            import_response = self.post_import(
                {
                    **response.context["form"].initial,
                    "from_index": 0,
                    "to_index": 1,
                    "permanent": True,
                }
            )

            self.assertTemplateUsed(
                import_response, "wagtailredirects/import_summary.html"
            )
            self.assertEqual(Redirect.objects.all().count(), 2)

    @override_settings(WAGTAIL_REDIRECTS_FILE_STORAGE="cache")
    def test_import_xlsx_with_cache_store_engine(self):
        f = f"{TEST_ROOT}/files/example.xlsx"
        (_, filename) = os.path.split(f)

        with open(f, "rb") as infile:
            upload_file = SimpleUploadedFile(filename, infile.read())

            self.assertEqual(Redirect.objects.all().count(), 0)

            response = self.post(
                {
                    "import_file": upload_file,
                }
            )

            self.assertTemplateUsed(response, "wagtailredirects/confirm_import.html")

            import_response = self.post_import(
                {
                    **response.context["form"].initial,
                    "from_index": 0,
                    "to_index": 1,
                    "permanent": True,
                },
                follow=True,
            )

            self.assertTemplateUsed(import_response, "wagtailredirects/index.html")
            self.assertEqual(Redirect.objects.all().count(), 3)

    @override_settings(WAGTAIL_REDIRECTS_FILE_STORAGE="cache")
    def test_process_validation_works_when_using_plaintext_files_and_cache(self):
        f = f"{TEST_ROOT}/files/example.csv"
        (_, filename) = os.path.split(f)

        with open(f, "rb") as infile:
            upload_file = SimpleUploadedFile(filename, infile.read())

            self.assertEqual(Redirect.objects.all().count(), 0)

            response = self.post(
                {
                    "import_file": upload_file,
                }
            )

            self.assertTemplateUsed(response, "wagtailredirects/confirm_import.html")

            import_response = self.post_import(
                {
                    **response.context["form"].initial,
                    "permanent": True,
                }
            )

            self.assertTemplateUsed(
                import_response, "wagtailredirects/confirm_import.html"
            )



from io import StringIO

from django.contrib.contenttypes.models import ContentType
from django.core import management
from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.test import TestCase
from django.utils.functional import SimpleLazyObject

from wagtail.blocks import StreamValue, StructValue
from wagtail.documents import get_document_model
from wagtail.documents.tests.utils import get_test_document_file
from wagtail.images import get_image_model
from wagtail.images.tests.utils import get_test_image_file
from wagtail.models import Page, ReferenceIndex
from wagtail.rich_text import RichText
from wagtail.test.testapp.models import (
    Advert,
    AdvertWithCustomUUIDPrimaryKey,
    EventPage,
    EventPageCarouselItem,
    EventPageRelatedLink,
    EventPageSpeaker,
    GenericSnippetNoFieldIndexPage,
    GenericSnippetNoIndexPage,
    GenericSnippetPage,
    HeadCountRelatedModelUsingPK,
    ModelWithNullableParentalKey,
    VariousOnDeleteModel,
)


class TestCreateOrUpdateForObject(TestCase):
    def setUp(self):
        with self.captureOnCommitCallbacks(execute=True):
            image_model = get_image_model()
            self.image_content_type = ContentType.objects.get_for_model(image_model)

            self.test_feed_image = image_model.objects.create(
                title="Test feed image",
                file=get_test_image_file(),
            )
            self.test_image_1 = image_model.objects.create(
                title="Test image 1",
                file=get_test_image_file(),
            )
            self.test_image_2 = image_model.objects.create(
                title="Test image 2",
                file=get_test_image_file(),
            )

            # Add event page
            self.event_page = EventPage(
                title="Event page",
                slug="event-page",
                location="the moon",
                audience="public",
                cost="free",
                date_from="2001-01-01",
                feed_image=self.test_feed_image,
            )
            self.event_page.carousel_items = [
                EventPageCarouselItem(
                    caption="1234567", image=self.test_image_1, sort_order=1
                ),
                EventPageCarouselItem(
                    caption="7654321", image=self.test_image_2, sort_order=2
                ),
                EventPageCarouselItem(
                    caption="abcdefg", image=self.test_image_1, sort_order=3
                ),
            ]
            self.root_page = Page.objects.get(id=2)
            self.root_page.add_child(instance=self.event_page)

        self.expected_references = {
            (
                self.image_content_type.id,
                str(self.test_feed_image.pk),
                "feed_image",
                "feed_image",
            ),
            (
                self.image_content_type.id,
                str(self.test_image_1.pk),
                "carousel_items.item.image",
                f"carousel_items.{self.event_page.carousel_items.get(sort_order=1).id}.image",
            ),
            (
                self.image_content_type.id,
                str(self.test_image_2.pk),
                "carousel_items.item.image",
                f"carousel_items.{self.event_page.carousel_items.get(sort_order=2).id}.image",
            ),
            (
                self.image_content_type.id,
                str(self.test_image_1.pk),
                "carousel_items.item.image",
                f"carousel_items.{self.event_page.carousel_items.get(sort_order=3).id}.image",
            ),
        }

    def test(self):
        self.assertSetEqual(
            set(
                ReferenceIndex.get_references_for_object(self.event_page).values_list(
                    "to_content_type", "to_object_id", "model_path", "content_path"
                )
            ),
            self.expected_references,
        )

    def test_update(self):
        reference_to_keep = ReferenceIndex.objects.get(
            base_content_type=ReferenceIndex._get_base_content_type(self.event_page),
            content_type=ContentType.objects.get_for_model(self.event_page),
            content_path="feed_image",
        )
        reference_to_remove = ReferenceIndex.objects.create(
            base_content_type=ReferenceIndex._get_base_content_type(self.event_page),
            content_type=ContentType.objects.get_for_model(self.event_page),
            object_id=self.event_page.pk,
            to_content_type=self.image_content_type,
            to_object_id=self.test_image_1.pk,
            model_path="hero_image",  # Field doesn't exist
            content_path="hero_image",
            content_path_hash=ReferenceIndex._get_content_path_hash("hero_image"),
        )

        ReferenceIndex.create_or_update_for_object(self.event_page)

        # Check that the record for the reference to be kept has been preserved/reused
        self.assertTrue(ReferenceIndex.objects.filter(id=reference_to_keep.id).exists())

        # Check that the record for the reference to be removed has been deleted
        self.assertFalse(
            ReferenceIndex.objects.filter(id=reference_to_remove.id).exists()
        )

        # Check that the current stored references are correct
        self.assertSetEqual(
            set(
                ReferenceIndex.get_references_for_object(self.event_page).values_list(
                    "to_content_type", "to_object_id", "model_path", "content_path"
                )
            ),
            {
                (
                    self.image_content_type.id,
                    str(self.test_feed_image.pk),
                    "feed_image",
                    "feed_image",
                ),
                (
                    self.image_content_type.id,
                    str(self.test_image_1.pk),
                    "carousel_items.item.image",
                    f"carousel_items.{self.event_page.carousel_items.get(sort_order=1).id}.image",
                ),
                (
                    self.image_content_type.id,
                    str(self.test_image_2.pk),
                    "carousel_items.item.image",
                    f"carousel_items.{self.event_page.carousel_items.get(sort_order=2).id}.image",
                ),
                (
                    self.image_content_type.id,
                    str(self.test_image_1.pk),
                    "carousel_items.item.image",
                    f"carousel_items.{self.event_page.carousel_items.get(sort_order=3).id}.image",
                ),
            },
        )

    def test_saving_base_model_does_not_remove_references(self):
        with self.captureOnCommitCallbacks(execute=True):
            page = Page.objects.get(pk=self.event_page.pk)
            page.save()
        self.assertSetEqual(
            set(
                ReferenceIndex.get_references_for_object(self.event_page).values_list(
                    "to_content_type", "to_object_id", "model_path", "content_path"
                )
            ),
            self.expected_references,
        )

    def test_null_parental_key(self):
        with self.captureOnCommitCallbacks(execute=True):
            obj = ModelWithNullableParentalKey(
                content="""<p><a linktype="page" id="%d">event page</a></p>"""
                % self.event_page.id
            )
            obj.save()

        # Models with a ParentalKey are not considered indexable - references are recorded against the parent model
        # instead. Since the ParentalKey is null here, no reference will be recorded.
        refs = ReferenceIndex.get_references_to(self.event_page)
        self.assertEqual(refs.count(), 0)

    def test_lazy_parental_key(self):
        with self.captureOnCommitCallbacks(execute=True):
            event_page_related_link = EventPageRelatedLink()
            # The parent model is a lazy object
            event_page_related_link.page = SimpleLazyObject(lambda: self.event_page)
            event_page_related_link.link_page = self.root_page
            event_page_related_link.save()
        refs = ReferenceIndex.get_references_to(self.root_page)
        self.assertEqual(refs.count(), 1)

    def test_generic_foreign_key(self):
        with self.captureOnCommitCallbacks(execute=True):
            page1 = GenericSnippetPage(
                title="generic snippet page", snippet_content_object=self.event_page
            )
            self.root_page.add_child(instance=page1)
            page2 = GenericSnippetPage(
                title="generic snippet page", snippet_content_object=None
            )
            self.root_page.add_child(instance=page2)

        refs = ReferenceIndex.get_references_to(self.event_page)
        self.assertEqual(refs.count(), 1)

    def test_model_index_ignore_generic_foreign_key(self):
        with self.captureOnCommitCallbacks(execute=True):
            page1 = GenericSnippetNoIndexPage(
                title="generic snippet page", snippet_content_object=self.event_page
            )
            self.root_page.add_child(instance=page1)
            page2 = GenericSnippetNoIndexPage(
                title="generic snippet page", snippet_content_object=None
            )
            self.root_page.add_child(instance=page2)

        # There should be no references
        refs = ReferenceIndex.get_references_to(self.event_page)
        self.assertEqual(refs.count(), 0)

    def test_model_field_index_ignore_generic_foreign_key(self):
        with self.captureOnCommitCallbacks(execute=True):
            content_type = ContentType.objects.get_for_model(self.event_page)
            page1 = GenericSnippetNoFieldIndexPage(
                title="generic snippet page",
                snippet_content_type_nonindexed=content_type,
            )
            self.root_page.add_child(instance=page1)
            page2 = GenericSnippetNoFieldIndexPage(
                title="generic snippet page", snippet_content_type_nonindexed=None
            )
            self.root_page.add_child(instance=page2)

        # There should be no references
        refs = ReferenceIndex.get_references_to(content_type)
        self.assertEqual(refs.count(), 0)

    def test_model_with_uuid_primary_key(self):
        refs = ReferenceIndex.get_references_to(self.event_page)
        self.assertEqual(refs.count(), 0)

        with self.captureOnCommitCallbacks(execute=True):
            AdvertWithCustomUUIDPrimaryKey.objects.create(
                text="An advertisement",
                page=self.event_page,
            )

        refs = ReferenceIndex.get_references_to(self.event_page)
        self.assertEqual(refs.count(), 1)

    def test_rebuild_references_index_no_verbosity(self):
        stdout = StringIO()
        management.call_command(
            "rebuild_references_index",
            verbosity=0,
            stdout=stdout,
        )
        self.assertFalse(stdout.getvalue())

    def test_show_references_index(self):
        stdout = StringIO()
        management.call_command(
            "show_references_index",
            stdout=stdout,
        )
        self.assertIn(" 3  wagtail.images.models.Image", stdout.getvalue())
        self.assertIn(" 4  wagtail.test.testapp.models.EventPage", stdout.getvalue())

    def test_inline_custom_pk_model(self):
        related_page = EventPage(
            title="Related page",
            slug="related-page",
            location="the moon",
            audience="public",
            cost="free",
            date_from="2025-03-21",
        )
        with self.captureOnCommitCallbacks(execute=True):
            self.root_page.add_child(instance=related_page)

        refs = ReferenceIndex.get_references_to(related_page)
        self.assertEqual(refs.count(), 0)

        with self.captureOnCommitCallbacks(execute=True):
            HeadCountRelatedModelUsingPK.objects.create(
                head_count=1234,
                event_page=self.event_page,
                related_page=related_page,
            )

        refs = ReferenceIndex.get_references_to(related_page)
        self.assertEqual(refs.count(), 1)


class TestDescribeOnDelete(TestCase):
    fixtures = ["test.json"]

    @classmethod
    def setUpTestData(cls):
        management.call_command("rebuild_references_index", stdout=StringIO())

    def setUp(self):
        field = VariousOnDeleteModel._meta.get_field("stream_field")
        advertisement_content = field.stream_block.child_blocks["advertisement_content"]
        captioned_advert = advertisement_content.child_blocks["captioned_advert"]

        self.advert = Advert.objects.create(text="An advertisement")
        self.advert_uuid = AdvertWithCustomUUIDPrimaryKey.objects.create(
            text="A UUID advertisement"
        )

        self.page = EventPage.objects.first()
        page_link = f'<p>Link to <a id="{self.page.id}" linktype="page">a page</a></p>'

        self.image = get_image_model().objects.create(
            title="My image",
            file=get_test_image_file(),
        )

        self.document = get_document_model().objects.create(
            title="My document",
            file=get_test_document_file(),
        )

        # Each case is a tuple of (
        #   VariousOnDeleteModel init kwargs,
        #   referred object,
        #   expected field description,
        #   expected on delete description,
        # )
        self.cases = [
            # References from ForeignKey
            (
                {"text": "on_delete=CASCADE", "on_delete_cascade": self.advert},
                self.advert,
                "On delete cascade",
                "the various on delete model will also be deleted",
            ),
            (
                {"text": "on_delete=PROTECT", "on_delete_protect": self.advert},
                self.advert,
                "On delete protect",
                "prevents deletion",
            ),
            (
                {"text": "on_delete=RESTRICT", "on_delete_restrict": self.advert},
                self.advert,
                "On delete restrict",
                "may prevent deletion",
            ),
            (
                {"text": "on_delete=SET_NULL", "on_delete_set_null": self.advert},
                self.advert,
                "On delete set null",
                "will unset the reference",
            ),
            (
                {"text": "on_delete=SET_DEFAULT", "on_delete_set_default": self.advert},
                self.advert,
                "On delete set default",
                "will be set to the default various on delete model",
            ),
            (
                {"text": "on_delete=SET", "on_delete_set": self.advert},
                self.advert,
                "On delete set",
                "will be set to a various on delete model specified by the system",
            ),
            (
                {"text": "on_delete=DO_NOTHING", "on_delete_do_nothing": self.advert},
                self.advert,
                "On delete do nothing",
                "will do nothing",
            ),
            # References from GenericForeignKey
            (
                {"text": "GenericForeignKey", "content_object": self.advert_uuid},
                self.advert_uuid,
                "Content object",
                "will unset the reference",
            ),
            # References from RichTextField
            (
                {"text": "RichTextField model field", "rich_text": page_link},
                self.page,
                "Rich text",
                "will unset the reference",
            ),
            (
                {
                    "text": "deep RichTextBlock",
                    "stream_field": StreamValue(
                        field.stream_block,
                        [
                            (
                                "advertisement_content",
                                StreamValue(
                                    advertisement_content,
                                    [
                                        (
                                            "rich_text",
                                            RichText(page_link),
                                        )
                                    ],
                                ),
                            )
                        ],
                    ),
                },
                self.page,
                "Stream field → Advertisement content → Rich text",
                "will unset the reference",
            ),
            # References from StreamField
            (
                {
                    "text": "deep SnippetChooserBlock",
                    "stream_field": StreamValue(
                        field.stream_block,
                        [
                            (
                                "advertisement_content",
                                StreamValue(
                                    advertisement_content,
                                    [
                                        (
                                            "captioned_advert",
                                            StructValue(
                                                captioned_advert,
                                                [
                                                    ("advert", self.advert),
                                                    ("caption", "Deep text"),
                                                ],
                                            ),
                                        )
                                    ],
                                ),
                            )
                        ],
                    ),
                },
                self.advert,
                "Stream field → Advertisement content → Captioned advert",
                "will unset the reference",
            ),
            (
                {
                    "text": "ImageChooserBlock",
                    "stream_field": StreamValue(
                        field.stream_block, [("image", self.image)]
                    ),
                },
                self.image,
                "Stream field → Image",
                "will unset the reference",
            ),
            (
                {
                    "text": "DocumentChooserBlock",
                    "stream_field": StreamValue(
                        field.stream_block, [("document", self.document)]
                    ),
                },
                self.document,
                "Stream field → Document",
                "will unset the reference",
            ),
        ]

    def test_describe_source_field_and_on_delete(self):
        for (
            init_kwargs,
            referred_object,
            field_description,
            on_delete_description,
        ) in self.cases:
            with self.subTest(test=init_kwargs["text"]):
                # Explicitly pass None to this field so that it is not set to
                # the default value for test cases other than the SET_DEFAULT case
                if "on_delete_set_default" not in init_kwargs:
                    init_kwargs["on_delete_set_default"] = None

                with self.captureOnCommitCallbacks(execute=True):
                    obj = VariousOnDeleteModel.objects.create(**init_kwargs)
                usage = ReferenceIndex.get_references_to(
                    referred_object
                ).group_by_source_object()
                referrer, references = usage[0]
                reference = references[0]

                self.assertIs(usage.is_protected, "on_delete_protect" in init_kwargs)
                self.assertEqual(usage.count(), 1)
                self.assertEqual(referrer, obj)
                self.assertEqual(len(references), 1)
                self.assertEqual(reference.describe_source_field(), field_description)
                self.assertEqual(reference.describe_on_delete(), on_delete_description)
                obj.delete()

    def test_describe_source_field_and_on_delete_parental_key(self):
        # The test fixtures contain two references to the advert:
        # 1. One advert placement on the home page
        # 2. One advert placement on the Christmas page
        advert = Advert.objects.first()
        usage = ReferenceIndex.get_references_to(advert).group_by_source_object()
        self.assertEqual(usage.count(), 2)
        for _, references in usage:
            reference = references[0]
            self.assertEqual(reference.describe_source_field(), "Advert")
            self.assertEqual(
                reference.describe_on_delete(),
                "the advert placement will also be deleted",
            )

    def test_parental_key_with_related_query_name(self):
        # EventPageSpeaker has a ParentalKey to EventPage with a
        # related_query_name that is different from the related_name, which
        # causes a mismatch between the recorded model_path and the way we
        # introspect the source field.
        root_page = Page.objects.get(id=2)
        speaker = EventPageSpeaker(
            first_name="Willie", last_name="Wagtail", link_page=root_page
        )
        event_page = EventPage.objects.first()
        event_page.speakers.add(speaker)
        with self.captureOnCommitCallbacks(execute=True):
            event_page.save()
        refs = ReferenceIndex.get_references_to(root_page)
        self.assertEqual(refs.count(), 1)
        self.assertEqual(refs[0].describe_source_field(), "Link page")
        self.assertEqual(
            refs[0].describe_on_delete(),
            "the event page speaker will also be deleted",
        )

    def test_nonexistent_field(self):
        # Simulate a situation where the field does not exist on the model
        # (e.g. due to a stale reference after the field was removed or renamed)
        reference = ReferenceIndex.objects.create(
            base_content_type=ReferenceIndex._get_base_content_type(self.advert),
            content_type=ContentType.objects.get_for_model(self.advert),
            object_id=self.advert.pk,
            to_content_type=ContentType.objects.get_for_model(self.page),
            to_object_id=self.page.pk,
            model_path="nonexistent_field",
            content_path="some_path",
            content_path_hash=ReferenceIndex._get_content_path_hash("some_path"),
        )
        with self.assertRaises(FieldDoesNotExist):
            reference.describe_source_field()


class TestBulkFetch(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.image_model = get_image_model()
        cls.document_model = get_document_model()

        cls.test_image_1 = cls.image_model.objects.create(
            title="Test image 1",
            file=get_test_image_file(),
        )
        cls.test_image_2 = cls.image_model.objects.create(
            title="Test image 2",
            file=get_test_image_file(),
        )
        cls.test_image_3 = cls.image_model.objects.create(
            title="Test image 3",
            file=get_test_image_file(),
        )
        cls.test_document_1 = cls.document_model.objects.create(
            title="Test document 1",
            file=get_test_document_file(),
        )
        cls.test_document_2 = cls.document_model.objects.create(
            title="Test document 2",
            file=get_test_document_file(),
        )

        cls.event_page = EventPage(
            title="Event page",
            slug="event-page",
            location="the moon",
            audience="public",
            cost="free",
            date_from="2001-01-01",
            feed_image=cls.test_image_1,
        )
        cls.event_page.carousel_items = [
            EventPageCarouselItem(
                caption="hello image 1", image=cls.test_image_1, sort_order=1
            ),
            EventPageCarouselItem(
                caption="hi image 2", image=cls.test_image_2, sort_order=2
            ),
            EventPageCarouselItem(
                caption="it's image 1 again", image=cls.test_image_1, sort_order=3
            ),
            EventPageCarouselItem(
                caption="now a doc", link_document=cls.test_document_2, sort_order=4
            ),
        ]
        cls.root_page = Page.objects.get(id=2)

        with cls.captureOnCommitCallbacks(execute=True):
            cls.root_page.add_child(instance=cls.event_page)
            cls.event_page.save()

        cls.carousel_items = list(cls.event_page.carousel_items.order_by("sort_order"))

    def test_get_references_to_in_bulk_with_empty_queryset_or_list(self):
        with self.assertNumQueries(0):
            refs = ReferenceIndex.get_references_to_in_bulk(
                get_image_model().objects.none()
            )
            self.assertIsInstance(refs, models.QuerySet)
            self.assertEqual(len(refs), 0)

        with self.assertNumQueries(0):
            refs = ReferenceIndex.get_references_to_in_bulk([])
            self.assertIsInstance(refs, models.QuerySet)
            self.assertEqual(len(refs), 0)

    def test_get_references_to_in_bulk_with_queryset(self):
        with self.assertNumQueries(1):
            refs = ReferenceIndex.get_references_to_in_bulk(
                get_image_model().objects.filter(
                    pk__in=[
                        self.test_image_1.pk,
                        self.test_image_2.pk,
                        self.test_image_3.pk,
                    ]
                )
            )
            self.assertIsInstance(refs, models.QuerySet)
            self.assertEqual(len(refs), 4)

        simplified_refs = {
            (
                ref.to_content_type_id,
                ref.to_object_id,
                ref.base_content_type_id,
                ref.object_id,
                ref.content_path,
            )
            for ref in refs
        }
        page_ct = ContentType.objects.get_for_model(Page)
        image_ct = ContentType.objects.get_for_model(self.image_model)
        page_id = str(self.event_page.pk)
        self.assertEqual(
            simplified_refs,
            {
                (
                    image_ct.pk,
                    str(self.test_image_1.pk),
                    page_ct.pk,
                    page_id,
                    f"carousel_items.{self.carousel_items[0].pk}.image",
                ),
                (
                    image_ct.pk,
                    str(self.test_image_1.pk),
                    page_ct.pk,
                    page_id,
                    f"carousel_items.{self.carousel_items[2].pk}.image",
                ),
                (
                    image_ct.pk,
                    str(self.test_image_1.pk),
                    page_ct.pk,
                    page_id,
                    "feed_image",
                ),
                (
                    image_ct.pk,
                    str(self.test_image_2.pk),
                    page_ct.pk,
                    page_id,
                    f"carousel_items.{self.carousel_items[1].pk}.image",
                ),
            },
        )

    def test_get_references_to_in_bulk_with_list_of_mixed_instances(self):
        with self.assertNumQueries(1):
            refs = ReferenceIndex.get_references_to_in_bulk(
                [
                    self.test_image_1,
                    self.test_image_2,
                    self.test_document_1,
                    self.test_document_2,
                ]
            )
            self.assertIsInstance(refs, models.QuerySet)
            self.assertEqual(len(refs), 5)

        simplified_refs = {
            (
                ref.to_content_type_id,
                ref.to_object_id,
                ref.base_content_type_id,
                ref.object_id,
                ref.content_path,
            )
            for ref in refs
        }
        page_ct = ContentType.objects.get_for_model(Page)
        image_ct = ContentType.objects.get_for_model(self.image_model)
        document_ct = ContentType.objects.get_for_model(self.document_model)
        page_id = str(self.event_page.pk)
        self.assertEqual(
            simplified_refs,
            {
                (
                    image_ct.pk,
                    str(self.test_image_1.pk),
                    page_ct.pk,
                    page_id,
                    f"carousel_items.{self.carousel_items[0].pk}.image",
                ),
                (
                    image_ct.pk,
                    str(self.test_image_1.pk),
                    page_ct.pk,
                    page_id,
                    f"carousel_items.{self.carousel_items[2].pk}.image",
                ),
                (
                    image_ct.pk,
                    str(self.test_image_1.pk),
                    page_ct.pk,
                    page_id,
                    "feed_image",
                ),
                (
                    image_ct.pk,
                    str(self.test_image_2.pk),
                    page_ct.pk,
                    page_id,
                    f"carousel_items.{self.carousel_items[1].pk}.image",
                ),
                (
                    document_ct.pk,
                    str(self.test_document_2.pk),
                    page_ct.pk,
                    page_id,
                    f"carousel_items.{self.carousel_items[3].pk}.link_document",
                ),
            },
        )

    def test_get_grouped_references_to_in_bulk_with_empty_queryset_or_list(self):
        with self.assertNumQueries(0):
            refs = ReferenceIndex.get_grouped_references_to_in_bulk(
                get_image_model().objects.none()
            )
            self.assertIsInstance(refs, dict)
            self.assertEqual(len(refs), 0)

        with self.assertNumQueries(0):
            refs = ReferenceIndex.get_grouped_references_to_in_bulk([])
            self.assertIsInstance(refs, dict)
            self.assertEqual(len(refs), 0)

    def test_get_grouped_references_to_in_bulk_with_queryset(self):
        # Two queries: one to get the references, one to evaluate the images so
        # that we can map each image to its references
        with self.assertNumQueries(2):
            refs = ReferenceIndex.get_grouped_references_to_in_bulk(
                get_image_model().objects.filter(
                    pk__in=[
                        self.test_image_1.pk,
                        self.test_image_2.pk,
                        self.test_image_3.pk,
                    ]
                )
            )
            self.assertIsInstance(refs, dict)
            self.assertEqual(len(refs), 3)

        expected_refs = {
            self.test_image_1: [
                (
                    self.event_page.page_ptr,
                    {
                        f"carousel_items.{self.carousel_items[0].pk}.image",
                        f"carousel_items.{self.carousel_items[2].pk}.image",
                        "feed_image",
                    },
                )
            ],
            self.test_image_2: [
                (
                    self.event_page.page_ptr,
                    {f"carousel_items.{self.carousel_items[1].pk}.image"},
                )
            ],
            self.test_image_3: [],
        }
        self.assertEqual(
            {
                item: [
                    (ref_group[0], {ref.content_path for ref in ref_group[1]})
                    for ref_group in ref_groups
                ]
                for item, ref_groups in refs.items()
            },
            expected_refs,
        )

    def test_get_grouped_references_to_in_bulk_with_list_of_mixed_instances(self):
        # We only need one query to get the references, as the images and
        # documents are already loaded in memory.
        with self.assertNumQueries(1):
            refs = ReferenceIndex.get_grouped_references_to_in_bulk(
                [
                    self.test_image_1,
                    self.test_image_2,
                    self.test_document_1,
                    self.test_document_2,
                ]
            )
            self.assertIsInstance(refs, dict)
            self.assertEqual(len(refs), 4)

        expected_refs = {
            self.test_image_1: [
                (
                    self.event_page.page_ptr,
                    {
                        f"carousel_items.{self.carousel_items[0].pk}.image",
                        f"carousel_items.{self.carousel_items[2].pk}.image",
                        "feed_image",
                    },
                )
            ],
            self.test_image_2: [
                (
                    self.event_page.page_ptr,
                    {f"carousel_items.{self.carousel_items[1].pk}.image"},
                )
            ],
            self.test_document_1: [],
            self.test_document_2: [
                (
                    self.event_page.page_ptr,
                    {f"carousel_items.{self.carousel_items[3].pk}.link_document"},
                )
            ],
        }

        self.assertEqual(
            {
                item: [
                    (ref_group[0], {ref.content_path for ref in ref_group[1]})
                    for ref_group in ref_groups
                ]
                for item, ref_groups in refs.items()
            },
            expected_refs,
        )



from django.apps import AppConfig
from django.db.models import ForeignKey
from django.utils.translation import gettext_lazy as _

from . import get_document_model


class WagtailDocsAppConfig(AppConfig):
    name = "wagtail.documents"
    label = "wagtaildocs"
    verbose_name = _("Wagtail documents")
    default_auto_field = "django.db.models.AutoField"

    def ready(self):
        from wagtail.documents.signal_handlers import register_signal_handlers

        register_signal_handlers()

        Document = get_document_model()

        from wagtail.admin.ui.fields import register_display_class

        from .components import DocumentDisplay

        register_display_class(ForeignKey, to=Document, display_class=DocumentDisplay)

        from wagtail.models.reference_index import ReferenceIndex

        ReferenceIndex.register_model(Document)



# Generated by Django 4.2.11 on 2024-04-25 15:51

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('wagtailcore', '0093_uploadedfile'),
        ('tests', '0036_complexdefaultstreampage'),
    ]

    operations = [
        migrations.CreateModel(
            name='TestPermissionedGenericSetting',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=100)),
                ('sensitive_email', models.EmailField(max_length=50)),
            ],
            options={
                'permissions': [('can_edit_sensitive_email_generic_setting', 'Can edit sensitive email generic setting.')],
            },
        ),
        migrations.AlterModelOptions(
            name='featurecompletetoy',
            options={'permissions': [('can_set_release_date', 'Can set release date')]},
        ),
        migrations.CreateModel(
            name='TestPermissionedSiteSetting',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=100)),
                ('sensitive_email', models.EmailField(max_length=50)),
                ('site', models.OneToOneField(editable=False, on_delete=django.db.models.deletion.CASCADE, to='wagtailcore.site')),
            ],
            options={
                'permissions': [('can_edit_sensitive_email_site_setting', 'Can edit sensitive email site setting.')],
            },
        ),
    ]



from django.test import RequestFactory, TestCase, override_settings

from wagtail import __version__
from wagtail.admin.views.home import UpgradeNotificationPanel
from wagtail.test.utils import WagtailTestUtils
from wagtail.users.models import UserProfile


class TestUpgradeNotificationPanel(WagtailTestUtils, TestCase):
    ATTR_UPGRADE_CHECK_LTS = "data-w-upgrade-lts-only-value"
    ATTR_CURRENT_VERSION = "data-w-upgrade-current-version-value"
    ATTR_DISMISSIBLE_ID = "data-w-dismissible-id-value"
    ATTR_LAST_DISMISSED_VALUE = "data-w-dismissible-value-param"
    DISMISSIBLE_ID = "last_upgrade_check"

    @classmethod
    def setUpTestData(cls):
        cls.panel = UpgradeNotificationPanel()
        cls.request_factory = RequestFactory()
        cls.user = cls.create_user(username="tester")
        cls.superuser = cls.create_superuser(username="supertester")
        cls.request = cls.request_factory.get("/")

    def test_get_upgrade_check_setting_default(self):
        self.assertTrue(self.panel.get_upgrade_check_setting())

    @override_settings(WAGTAIL_ENABLE_UPDATE_CHECK=False)
    def test_get_upgrade_check_setting_false(self):
        self.assertFalse(self.panel.get_upgrade_check_setting())

    @override_settings(WAGTAIL_ENABLE_UPDATE_CHECK="LTS")
    def test_get_upgrade_check_setting_LTS(self):
        self.assertEqual(self.panel.get_upgrade_check_setting(), "LTS")

    @override_settings(WAGTAIL_ENABLE_UPDATE_CHECK="lts")
    def test_get_upgrade_check_setting_lts(self):
        self.assertEqual(self.panel.get_upgrade_check_setting(), "lts")

    def test_upgrade_check_lts_only_default(self):
        self.assertFalse(self.panel.upgrade_check_lts_only())

    @override_settings(WAGTAIL_ENABLE_UPDATE_CHECK=False)
    def test_upgrade_check_lts_only_setting_true(self):
        self.assertFalse(self.panel.upgrade_check_lts_only())

    @override_settings(WAGTAIL_ENABLE_UPDATE_CHECK="LTS")
    def test_upgrade_check_lts_only_setting_LTS(self):
        self.assertTrue(self.panel.upgrade_check_lts_only())

    @override_settings(WAGTAIL_ENABLE_UPDATE_CHECK="lts")
    def test_upgrade_check_lts_only_setting_lts(self):
        self.assertTrue(self.panel.upgrade_check_lts_only())

    def test_render_html_normal_user(self):
        self.request.user = self.user
        parent_context = {"request": self.request}

        result = self.panel.render_html(parent_context)

        self.assertEqual(result, "")

    def test_render_html_superuser(self):
        self.request.user = self.superuser
        parent_context = {"request": self.request}

        result = self.panel.render_html(parent_context)
        soup = self.get_soup(result)
        controller = soup.select_one("[data-controller]")
        self.assertIsNotNone(controller)
        self.assertEqual(
            set(controller["data-controller"].split()),
            {"w-upgrade", "w-dismissible"},
        )
        self.assertFalse(controller.get(self.ATTR_UPGRADE_CHECK_LTS))
        self.assertEqual(
            controller.get(self.ATTR_DISMISSIBLE_ID),
            self.DISMISSIBLE_ID,
        )
        toggle = soup.select_one("[data-action='w-dismissible#toggle']")
        self.assertIsNotNone(toggle)
        self.assertEqual(toggle.get("aria-label"), "Close")
        self.assertIsNone(toggle.get(self.ATTR_LAST_DISMISSED_VALUE))

    @override_settings(WAGTAIL_ENABLE_UPDATE_CHECK=False)
    def test_render_html_setting_false(self):
        self.request.user = self.superuser
        parent_context = {"request": self.request}

        result = self.panel.render_html(parent_context)

        self.assertEqual(result, "")

    def test_render_html_setting_lts(self):
        self.request.user = self.superuser
        parent_context = {"request": self.request}
        setting_values = ["lts", "LTS"]
        for value in setting_values:
            with self.subTest(setting=value):
                with override_settings(WAGTAIL_ENABLE_UPDATE_CHECK=value):
                    result = self.panel.render_html(parent_context)

                soup = self.get_soup(result)
                controller = soup.select_one("[data-controller]")
                self.assertIsNotNone(controller)
                self.assertEqual(
                    set(controller["data-controller"].split()),
                    {"w-upgrade", "w-dismissible"},
                )
                self.assertEqual(
                    controller.get(self.ATTR_UPGRADE_CHECK_LTS),
                    "true",
                )
                self.assertEqual(
                    controller.get(self.ATTR_DISMISSIBLE_ID),
                    self.DISMISSIBLE_ID,
                )
                toggle = soup.select_one("[data-action='w-dismissible#toggle']")
                self.assertIsNotNone(toggle)
                self.assertEqual(toggle.get("aria-label"), "Close")
                self.assertIsNone(toggle.get(self.ATTR_LAST_DISMISSED_VALUE))

    def test_render_html_dismissed_version(self):
        profile = UserProfile.get_for_user(self.superuser)
        profile.dismissibles.update({self.DISMISSIBLE_ID: "6.2.2"})
        profile.save()
        self.request.user = self.superuser
        parent_context = {"request": self.request}

        result = self.panel.render_html(parent_context)
        soup = self.get_soup(result)

        controller = soup.select_one("[data-controller='w-upgrade w-dismissible']")
        self.assertIsNotNone(controller)

        self.assertEqual(
            controller.get(self.ATTR_DISMISSIBLE_ID),
            self.DISMISSIBLE_ID,
        )
        self.assertEqual(
            controller.get(self.ATTR_CURRENT_VERSION),
            __version__,
        )
        toggle = soup.select_one("[data-action='w-dismissible#toggle']")
        self.assertIsNotNone(toggle)
        self.assertEqual(toggle.get("aria-label"), "Close")
        self.assertEqual(
            toggle.get(self.ATTR_LAST_DISMISSED_VALUE),
            "6.2.2",
        )



from django.contrib.auth.models import Group
from django.urls import reverse
from django.utils.functional import cached_property
from django.utils.translation import gettext
from django.utils.translation import gettext_lazy as _

from wagtail import hooks
from wagtail.admin.ui.tables import TitleColumn
from wagtail.admin.utils import set_query_params
from wagtail.admin.views import generic
from wagtail.admin.viewsets.model import ModelViewSet
from wagtail.admin.widgets.button import HeaderButton
from wagtail.users.forms import GroupForm, GroupPagePermissionFormSet

_permission_panel_classes = None


def get_permission_panel_classes():
    global _permission_panel_classes
    if _permission_panel_classes is None:
        _permission_panel_classes = [GroupPagePermissionFormSet]
        for fn in hooks.get_hooks("register_group_permission_panel"):
            _permission_panel_classes.append(fn())

    return _permission_panel_classes


class PermissionPanelFormsMixin:
    def get_permission_panel_form_kwargs(self, cls):
        kwargs = {}

        if self.request.method in ("POST", "PUT"):
            kwargs.update(
                {
                    "data": self.request.POST,
                    "files": self.request.FILES,
                }
            )

        if hasattr(self, "object"):
            kwargs.update({"instance": self.object})

        return kwargs

    def get_permission_panel_forms(self):
        return [
            cls(**self.get_permission_panel_form_kwargs(cls))
            for cls in get_permission_panel_classes()
        ]

    def process_form(self):
        form = self.get_form()
        permission_panels = self.get_permission_panel_forms()
        if form.is_valid() and all(panel.is_valid() for panel in permission_panels):
            response = self.form_valid(form)

            for panel in permission_panels:
                panel.save()

            return response
        else:
            return self.form_invalid(form)

    def get_context_data(self, **kwargs):
        if "permission_panels" not in kwargs:
            kwargs["permission_panels"] = self.get_permission_panel_forms()

        context = super().get_context_data(**kwargs)

        # Add js/css media from the formsets to the existing media
        for panel in context["permission_panels"]:
            context["media"] += panel.media

        return context


class IndexView(generic.IndexView):
    page_title = _("Groups")
    add_item_label = _("Add a group")
    search_box_placeholder = _("Search groups")
    search_fields = ["name"]
    context_object_name = "groups"
    paginate_by = 20

    columns = [
        TitleColumn(
            "name",
            label=_("Name"),
            sort_key="name",
            url_name="wagtailusers_groups:edit",
        ),
    ]


class CreateView(PermissionPanelFormsMixin, generic.CreateView):
    page_title = _("Add group")
    success_message = _("Group '%(object)s' created.")

    def post(self, request, *args, **kwargs):
        """
        Handle POST requests: instantiate a form instance with the passed
        POST variables and then check if it's valid.
        """
        # Create an object now so that the permission panel forms have something to link them against
        self.object = Group()
        return self.process_form()


class EditView(PermissionPanelFormsMixin, generic.EditView):
    success_message = _("Group '%(object)s' updated.")
    error_message = _("The group could not be saved due to errors.")
    context_object_name = "group"

    @cached_property
    def header_buttons(self):
        return [
            HeaderButton(
                gettext("View users in this group"),
                url=set_query_params(
                    reverse("wagtailusers_users:index"),
                    {"group": self.object.pk},
                ),
                icon_name="user",
            )
        ]

    def post(self, request, *args, **kwargs):
        """
        Handle POST requests: instantiate a form instance with the passed
        POST variables and then check if it's valid.
        """
        self.object = self.get_object()
        return self.process_form()


class DeleteView(generic.DeleteView):
    success_message = _("Group '%(object)s' deleted.")
    page_title = _("Delete group")
    confirmation_message = _("Are you sure you want to delete this group?")


class GroupViewSet(ModelViewSet):
    icon = "group"
    model = Group
    ordering = ["name"]
    add_to_reference_index = False
    menu_name = "groups"
    menu_label = _("Groups")
    menu_order = 601
    add_to_settings_menu = True

    index_view_class = IndexView
    add_view_class = CreateView
    edit_view_class = EditView
    delete_view_class = DeleteView

    # Copying requires prepopulating the formsets with copies of the existing
    # permissions, which is not currently supported.
    copy_view_enabled = False

    template_prefix = "wagtailusers/groups/"

    def get_common_view_kwargs(self, **kwargs):
        return super().get_common_view_kwargs(
            **{
                "usage_url_name": None,
                **kwargs,
            }
        )

    def get_form_class(self, for_update=False):
        return GroupForm



# Generated by Django 1.11.6 on 2017-10-18 12:59
from django.db import migrations, models
import django.db.models.deletion
import taggit.managers
import wagtail.search.index


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("taggit", "0002_auto_20150616_2121"),
    ]

    operations = [
        migrations.CreateModel(
            name="Author",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("date_of_birth", models.DateField(null=True)),
            ],
            bases=(models.Model, wagtail.search.index.Indexed),
        ),
        migrations.CreateModel(
            name="Book",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("title", models.CharField(max_length=255)),
                ("publication_date", models.DateField()),
                ("number_of_pages", models.IntegerField()),
            ],
            bases=(models.Model, wagtail.search.index.Indexed),
        ),
        migrations.CreateModel(
            name="Character",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name="Novel",
            fields=[
                (
                    "book_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="searchtests.Book",
                    ),
                ),
                ("setting", models.CharField(max_length=255)),
                (
                    "protagonist",
                    models.OneToOneField(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to="searchtests.Character",
                    ),
                ),
            ],
            bases=("searchtests.book",),
        ),
        migrations.CreateModel(
            name="ProgrammingGuide",
            fields=[
                (
                    "book_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="searchtests.Book",
                    ),
                ),
                (
                    "programming_language",
                    models.CharField(
                        choices=[
                            ("py", "Python"),
                            ("js", "JavaScript"),
                            ("rs", "Rust"),
                        ],
                        max_length=255,
                    ),
                ),
            ],
            bases=("searchtests.book",),
        ),
        migrations.AddField(
            model_name="book",
            name="authors",
            field=models.ManyToManyField(related_name="books", to="searchtests.Author"),
        ),
        migrations.AddField(
            model_name="book",
            name="tags",
            field=taggit.managers.TaggableManager(
                help_text="A comma-separated list of tags.",
                through="taggit.TaggedItem",
                to="taggit.Tag",
                verbose_name="Tags",
            ),
        ),
        migrations.AddField(
            model_name="character",
            name="novel",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="characters",
                to="searchtests.Novel",
            ),
        ),
    ]



from django.core.paginator import EmptyPage, PageNotAnInteger, Paginator
from django.template.response import TemplateResponse

from wagtail.models import Page

# To enable logging of search queries for use with the "Promoted search results" module
# <https://docs.wagtail.org/en/stable/reference/contrib/searchpromotions.html>
# uncomment the following line and the lines indicated in the search function
# (after adding wagtail.contrib.search_promotions to INSTALLED_APPS):

# from wagtail.contrib.search_promotions.models import Query


def search(request):
    search_query = request.GET.get("query", None)
    page = request.GET.get("page", 1)

    # Search
    if search_query:
        search_results = Page.objects.live().search(search_query)

        # To log this query for use with the "Promoted search results" module:

        # query = Query.get(search_query)
        # query.add_hit()

    else:
        search_results = Page.objects.none()

    # Pagination
    paginator = Paginator(search_results, 10)
    try:
        search_results = paginator.page(page)
    except PageNotAnInteger:
        search_results = paginator.page(1)
    except EmptyPage:
        search_results = paginator.page(paginator.num_pages)

    return TemplateResponse(
        request,
        "search/search.html",
        {
            "search_query": search_query,
            "search_results": search_results,
        },
    )
